#
# Copyright (c) 2021, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import tensorrt as trt
from .calibrator import Calibrator
from alfred.utils.log import logger

import sys
import os
import time
import os
import os.path as osp
import ctypes

try:
    import pycuda.driver as cuda
    # https://documen.tician.de/pycuda/driver.html
except ImportError as e:
    print(
        f'pycuda not installed, or import failed. inference on trt will be disabled. error: {e}')

TRT8 = 8
TRT7 = 7

TRT_LOGGER = trt.Logger(trt.Logger.INFO)
# TRT_LOGGER = trt.Logger()
trt.init_libnvinfer_plugins(TRT_LOGGER, namespace="")
# Allocate host and device buffers, and create a stream.


class HostDeviceMem(object):
    def __init__(self, host_mem, device_mem):
        self.host = host_mem
        self.device = device_mem

    def __str__(self):
        return "Host:\n" + str(self.host) + "\nDevice:\n" + str(self.device)

    def __repr__(self):
        return self.__str__()


def allocate_buffers(engine):
    inputs = []
    outputs = []
    bindings = []
    stream = cuda.Stream()
    for binding in engine:
        # <--------- the main diff to v2
        size = trt.volume(engine.get_binding_shape(binding))
        dtype = trt.nptype(engine.get_binding_dtype(binding))
        # Allocate host and device buffers
        host_mem = cuda.pagelocked_empty(size, dtype)
        device_mem = cuda.mem_alloc(host_mem.nbytes)
        # Append the device buffer to device bindings.
        bindings.append(int(device_mem))
        # Append to the appropriate list.
        if engine.binding_is_input(binding):
            inputs.append(HostDeviceMem(host_mem, device_mem))
        else:
            outputs.append(HostDeviceMem(host_mem, device_mem))
    return inputs, outputs, bindings, stream


def allocate_buffers_v2(engine):
    inputs = []
    outputs = []
    bindings = []
    stream = cuda.Stream()
    for binding in engine:
        size = abs(trt.volume(engine.get_binding_shape(binding))) * \
            engine.max_batch_size
        dtype = trt.nptype(engine.get_binding_dtype(binding))
        # Allocate host and device buffers
        host_mem = cuda.pagelocked_empty(size, dtype)
        device_mem = cuda.mem_alloc(host_mem.nbytes)
        # Append the device buffer to device bindings.
        bindings.append(int(device_mem))
        # Append to the appropriate list.
        if engine.binding_is_input(binding):
            inputs.append(HostDeviceMem(host_mem, device_mem))
        else:
            outputs.append(HostDeviceMem(host_mem, device_mem))
    return inputs, outputs, bindings, stream
    # return inputs, outputs, bindings


# do inference  multi outputs
def do_inference_v2(context, bindings, inputs, outputs, stream, input_tensor):
    # Transfer input data to the GPU.
    [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]

    # Run inference.
    context.execute_async_v2(bindings=bindings, stream_handle=stream.handle)
    # Transfer predictions back from the GPU.
    [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
    # Synchronize the stream
    stream.synchronize()
    # Return only the host outputs.
    return [out.host for out in outputs]


def allocate_buffers_v2_dynamic(engine, is_explicit_batch=False, input_shape=None):
    inputs = []
    outputs = []
    bindings = []
    stream = cuda.Stream()
    for binding in engine:
        dims = engine.get_binding_shape(binding)
        if dims[-1] == -1:
            assert (
                input_shape is not None
            ), "dynamic trt engine must specific input_shape"
            dims[-2], dims[-1] = input_shape
        size = abs(trt.volume(dims)) * engine.max_batch_size
        dtype = trt.nptype(engine.get_binding_dtype(binding))
        # Allocate host and device buffers
        host_mem = cuda.pagelocked_empty(size, dtype)
        device_mem = cuda.mem_alloc(host_mem.nbytes)
        # Append the device buffer to device bindings.
        bindings.append(int(device_mem))
        # Append to the appropriate list.
        if engine.binding_is_input(binding):
            inputs.append(HostDeviceMem(host_mem, device_mem))
        else:
            outputs.append(HostDeviceMem(host_mem, device_mem))
    return inputs, outputs, bindings, stream


def allocate_buffers_v2_dynamic_batch(engine, is_explicit_batch=False, max_batch=1):
    inputs = []
    outputs = []
    bindings = []
    stream = cuda.Stream()
    for binding in engine:
        dims = engine.get_binding_shape(binding)
        if dims[-1] == -1:
            assert (
                input_shape is not None
            ), "dynamic trt engine must specific input_shape"
            dims[-2], dims[-1] = input_shape
        size = trt.volume(dims) * engine.max_batch_size
        dtype = trt.nptype(engine.get_binding_dtype(binding))
        # Allocate host and device buffers
        host_mem = cuda.pagelocked_empty(size, dtype)
        device_mem = cuda.mem_alloc(host_mem.nbytes)
        # Append the device buffer to device bindings.
        bindings.append(int(device_mem))
        # Append to the appropriate list.
        if engine.binding_is_input(binding):
            inputs.append(HostDeviceMem(host_mem, device_mem))
        else:
            outputs.append(HostDeviceMem(host_mem, device_mem))
    return inputs, outputs, bindings, stream


def do_inference_v2_dynamic(context, bindings, inputs, outputs, stream, input_tensor):
    """
    this works for infer on dynamic engine, such as input dynamic
    """
    # Transfer input data to the GPU.
    [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]

    # Run inference.
    context.execute_async_v2(bindings=bindings, stream_handle=stream.handle)
    # Transfer predictions back from the GPU.
    [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
    # Synchronize the stream
    stream.synchronize()
    # Return only the host outputs.
    return [out.host for out in outputs]


def build_engine_onnx(
    model_file,
    engine_file,
    FP16=False,
    verbose=False,
    dynamic_input=False,
    batch_size=1,
    chw_shape=None,
):
    def get_engine():
        EXPLICIT_BATCH = 1 << (int)(
            trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
        # with trt.Builder(TRT_LOGGER) as builder, builder.create_network(EXPLICIT_BATCH) as network,builder.create_builder_config() as config, trt.OnnxParser(network,TRT_LOGGER) as parser:
        with trt.Builder(TRT_LOGGER) as builder, builder.create_network(
            EXPLICIT_BATCH
        ) as network, builder.create_builder_config() as config, trt.OnnxParser(
            network, TRT_LOGGER
        ) as parser:

            trt_version = int(trt.__version__[0])
            # Workspace size is the maximum amount of memory available to the builder while building an engine.

            if trt_version == TRT8:
                config.max_workspace_size = 6 << 30  # 2GB
            else:
                builder.max_workspace_size = 6 << 30  # 2GB

            if trt_version == TRT8:
                if FP16:
                    config.set_flag(trt.BuilderFlag.FP16)
            else:
                builder.fp16_mode = FP16

            with open(model_file, "rb") as model:
                parser.parse(model.read())
            if verbose:
                logger.info(">" * 50)
                for error in range(parser.num_errors):
                    logger.info(parser.get_error(error))

            # network.get_input(0).shape = [batch_size, 3, 800, 800]
            if chw_shape:
                network.get_input(0).shape = [batch_size, *chw_shape]

            if dynamic_input:
                profile = builder.create_optimization_profile()
                profile.set_shape(
                    "inputs", (1, 3, 800, 800), (8, 3,
                                                 800, 800), (64, 3, 800, 800)
                )
                config.add_optimization_profile(profile)

            # builder engine
            engine = None
            if trt_version == TRT8:
                engine = builder.build_engine(network, config)
            else:
                engine = builder.build_cuda_engine(network)

            if engine is not None:
                logger.info("[INFO] Completed creating Engine!")
                with open(engine_file, "wb") as f:
                    f.write(engine.serialize())
            else:
                logger.error("Create engine failed!")
            return engine

    if os.path.exists(engine_file):
        # If a serialized engine exists, use it instead of building an engine.
        logger.info("[INFO] Reading engine from file {}".format(engine_file))
        with open(engine_file, "rb") as f, trt.Runtime(TRT_LOGGER) as runtime:
            return runtime.deserialize_cuda_engine(f.read())
    else:
        return get_engine()


# int8 quant
def build_engine_onnx_v2(
    onnx_file_path="",
    engine_file_path="",
    fp16_mode=False,
    int8_mode=False,
    max_batch_size=1,
    calibration_stream=None,
    calibration_table_path="",
    save_engine=False,
    opt_params: dict = None,
):
    """Attempts to load a serialized engine if available, otherwise builds a new TensorRT engine and saves it."""

    def build_engine(max_batch_size, save_engine, opt_params=None):
        """Takes an ONNX file and creates a TensorRT engine to run inference with"""
        with trt.Builder(TRT_LOGGER) as builder, builder.create_network(
            1
        ) as network, builder.create_builder_config() as config, trt.OnnxParser(
            network, TRT_LOGGER
        ) as parser, builder.create_builder_config() as trt_config:

            trt_version = int(trt.__version__[0])
            # parse onnx model file
            if not os.path.exists(onnx_file_path):
                quit(f"[Error]ONNX file {onnx_file_path} not found")
            logger.info(
                f"[INFO] Loading ONNX file from path {onnx_file_path}...")
            with open(onnx_file_path, "rb") as model:
                logger.info("[INFO] Beginning ONNX file parsing")
                parser.parse(model.read())
                assert (
                    network.num_layers > 0
                ), "[Error] Failed to parse ONNX model. \
                            Please check if the ONNX model is compatible "
            logger.info("[INFO] Completed parsing of ONNX file")
            logger.info(
                f"[INFO] Building an engine from file {onnx_file_path}, this may take a while..."
            )

            # build trt engine

            # set optimize profile for dynamic inputs
            profile = builder.create_optimization_profile()
            if opt_params is not None:
                """
                opt_params = {
                    'input': [
                        [1, 3, 416, 502], # min shape
                        [1, 3, 416, 502], # opt shape
                        [1, 3, 416, 502], # max shape
                    ]
                }
                """
                logger.info('using opt_params: {}'.format(opt_params))
                for input_index, input_tensor_name in enumerate(opt_params.keys()):
                    min_shape = tuple(opt_params[input_tensor_name][0][:])
                    opt_shape = tuple(opt_params[input_tensor_name][1][:])
                    max_shape = tuple(opt_params[input_tensor_name][2][:])
                    profile.set_shape(
                        input_tensor_name, min_shape, opt_shape, max_shape
                    )
                    max_batch_size = max_shape[0]

                builder.max_batch_size = max_batch_size
            trt_config.add_optimization_profile(profile)

            if trt_version == TRT8:
                trt_config.max_workspace_size = 2 << 30  # 2GB
            else:
                builder.max_workspace_size = 2 << 30  # 2GB

            if trt_version == TRT8:
                if fp16_mode:
                    trt_config.set_flag(trt.BuilderFlag.FP16)
                    logger.info('enabled fp16 mode.')
            else:
                if fp16_mode:
                    builder.fp16_mode = fp16_mode
                    logger.info('enabled fp16 mode.')

            if int8_mode:
                if trt_version == TRT8:
                    trt_config.set_flag(trt.BuilderFlag.INT8)
                    logger.info('enabled int8 mode.')
                else:
                    builder.fp16_mode = fp16_mode

                assert (
                    calibration_stream
                ), "[Error] a calibration_stream should be provided for int8 mode"
                config.int8_calibrator = Calibrator(
                    calibration_stream, calibration_table_path
                )
                # builder.int8_calibrator  = Calibrator(calibration_stream, calibration_table_path)
                logger.info("Int8 mode enabled")

            engine = None
            if trt_version == TRT8:
                engine = builder.build_engine(network, trt_config)
            else:
                engine = builder.build_cuda_engine(network)

            if engine is None:
                logger.info("Failed to create the engine")
                return None
            logger.info("Completed creating the engine")
            if save_engine:
                with open(engine_file_path, "wb") as f:
                    f.write(engine.serialize())
            return engine

    if os.path.exists(engine_file_path):
        # If a serialized engine exists, load it instead of building a new one.
        logger.info(f"Reading engine from file {engine_file_path}")
        with open(engine_file_path, "rb") as f, trt.Runtime(TRT_LOGGER) as runtime:
            return runtime.deserialize_cuda_engine(f.read())
    else:
        return build_engine(max_batch_size, save_engine, opt_params=opt_params)


def load_engine_from_local(engine_file_path):
    if os.path.exists(engine_file_path):
        # If a serialized engine exists, load it instead of building a new one.
        logger.info(f"Reading engine from file {engine_file_path}")
        with open(engine_file_path, "rb") as f, trt.Runtime(TRT_LOGGER) as runtime:
            return runtime.deserialize_cuda_engine(f.read())
    else:
        logger.info("engine does not exist, please build it first.")
        exit(1)


def load_torchtrt_plugins():
    # ctypes.CDLL(osp.join(dir_path, 'libamirstan_plugin.so'))
    # suppose plugins lib installed into HOME
    lib_path = osp.join(
        osp.expanduser(
            "~"), "torchtrt_plugins/build/lib/libtorchtrt_plugins.so"
    )
    if os.path.exists(lib_path):
        ctypes.CDLL(lib_path)
    else:
        logger.warning(f"{lib_path} not found.")


def build_engine_onnx_v3(
    onnx_file_path="",
    fp16_mode=False,
    int8_mode=False,
    max_batch_size=1,
    calibration_stream=None,
    calibration_table_path="",
    save_engine=True,
):
    """
    this is deprecated.
    """
    engine_file_path = os.path.join(
        os.path.dirname(onnx_file_path),
        os.path.basename(onnx_file_path).replace(".onnx", ".engine"),
    )
    trt.init_libnvinfer_plugins(None, "")

    def build_engine(max_batch_size, save_engine):
        """Takes an ONNX file and creates a TensorRT engine to run inference with"""
        with trt.Builder(TRT_LOGGER) as builder, builder.create_network(
            1
        ) as network, builder.create_builder_config() as config, trt.OnnxParser(
            network, TRT_LOGGER
        ) as parser, builder.create_builder_config() as trt_config:

            trt_version = int(trt.__version__[0])
            # parse onnx model file
            if not os.path.exists(onnx_file_path):
                quit(f"[Error]ONNX file {onnx_file_path} not found")
            logger.info(f"Loading ONNX file from path {onnx_file_path}...")
            with open(onnx_file_path, "rb") as model:
                logger.info("Beginning ONNX file parsing")
                parser.parse(model.read())
                assert (
                    network.num_layers > 0
                ), "[Error] Failed to parse ONNX model. \
                            Please check if the ONNX model is compatible "
            logger.info("[INFO] Completed parsing of ONNX file")
            logger.info(
                f"Building an engine from file {onnx_file_path}, this may take a while..."
            )

            # build trt engine
            builder.max_batch_size = max_batch_size

            if trt_version == TRT8:
                trt_config.max_workspace_size = 2 << 30  # 2GB
            else:
                builder.max_workspace_size = 2 << 30  # 2GB

            if trt_version == TRT8:
                if fp16_mode:
                    trt_config.set_flag(trt.BuilderFlag.FP16)
            else:
                builder.fp16_mode = fp16_mode

            if int8_mode:
                if trt_version == TRT8:
                    trt_config.set_flag(trt.BuilderFlag.INT8)
                else:
                    builder.fp16_mode = fp16_mode

                assert (
                    calibration_stream
                ), "[Error] a calibration_stream should be provided for int8 mode"
                config.int8_calibrator = Calibrator(
                    calibration_stream, calibration_table_path
                )
                # builder.int8_calibrator  = Calibrator(calibration_stream, calibration_table_path)
                logger.info("Int8 mode enabled")

            engine = None
            if trt_version == TRT8:
                engine = builder.build_engine(network, trt_config)
            else:
                engine = builder.build_cuda_engine(network)

            if engine is None:
                logger.info("Failed to create the engine")
                return None
            logger.info("Completed creating the engine")
            if save_engine:
                logger.info(f"engine saved into: {engine_file_path}")
                with open(engine_file_path, "wb") as f:
                    f.write(engine.serialize())
            return engine

    if os.path.exists(engine_file_path):
        # If a serialized engine exists, load it instead of building a new one.
        logger.info(f"Reading engine from file {engine_file_path}")
        with open(engine_file_path, "rb") as f, trt.Runtime(TRT_LOGGER) as runtime:
            return runtime.deserialize_cuda_engine(f.read())
    else:
        return build_engine(max_batch_size, save_engine)


def check_engine(engine, input_shape=(608, 608), print=False):
    tensor_names_shape_dict = {}
    for binding in engine:
        dims = engine.get_binding_shape(binding)
        if dims[-1] == -1:
            assert (
                input_shape is not None
            ), "dynamic trt engine must specific input_shape"
            dims[-2], dims[-1] = input_shape
        size = trt.volume(dims) * engine.max_batch_size
        dtype = np.dtype(trt.nptype(engine.get_binding_dtype(binding)))

        if engine.binding_is_input(binding):
            tensor_names_shape_dict[binding] = {
                "shape": dims,
                "dtype": dtype,
                "is_input": True,
            }
            if print:
                print(f"[{binding}]: {dims}, {dtype.name}, [INPUT]")
        else:
            tensor_names_shape_dict[binding] = {
                "shape": dims,
                "dtype": dtype,
                "is_input": False,
            }
            if print:
                print(f"[{binding}]: {dims}, {dtype.name}, [OUTPUT]")
    return tensor_names_shape_dict



from .sliceable_dataset import SliceableDataset


class ConcatenatedDataset(SliceableDataset):
    """A sliceable version of :class:`chainer.datasets.ConcatenatedDataset`.

    Here is an example.

    >>> dataset_a = TupleDataset([0, 1, 2], [0, 1, 4])
    >>> dataset_b = TupleDataset([3, 4, 5], [9, 16, 25])
    >>>
    >>> dataset = ConcatenatedDataset(dataset_a, dataset_b)
    >>> dataset.slice[:, 0][:]  # [0, 1, 2, 3, 4, 5]

    Args:
        datasets: The underlying datasets.
            Each dataset should inherit
            :class:`~chainercv.chainer_experimental.datasets.sliceable.Sliceabledataset`
            and should have the same keys.
    """

    def __init__(self, *datasets):
        if len(datasets) == 0:
            raise ValueError('At least one dataset is required')
        self._datasets = datasets
        self._keys = datasets[0].keys
        for dataset in datasets[1:]:
            if not dataset.keys == self._keys:
                raise ValueError('All datasets should have the same keys')

    def __len__(self):
        return sum(len(dataset) for dataset in self._datasets)

    @property
    def keys(self):
        return self._keys

    def get_example_by_keys(self, index, key_indices):
        if index < 0:
            raise IndexError
        for dataset in self._datasets:
            if index < len(dataset):
                return dataset.get_example_by_keys(index, key_indices)
            index -= len(dataset)
        raise IndexError


import math

import torch
import torch.nn as nn

from .utils import weighted_loss, bbox_overlaps


@weighted_loss
def iou_loss(pred, target, eps=1e-6):
    """IoU loss.

    Computing the IoU loss between a set of predicted bboxes and target bboxes.
    The loss is calculated as negative log of IoU.

    Args:
        pred (torch.Tensor): Predicted bboxes of format (x1, y1, x2, y2),
            shape (n, 4).
        target (torch.Tensor): Corresponding gt bboxes, shape (n, 4).
        eps (float): Eps to avoid log(0).

    Return:
        torch.Tensor: Loss tensor.
    """
    ious = bbox_overlaps(pred, target, is_aligned=True).clamp(min=eps)
    loss = -ious.log()
    return loss


@weighted_loss
def bounded_iou_loss(pred, target, beta=0.2, eps=1e-3):
    """BIoULoss.

    This is an implementation of paper
    `Improving Object Localization with Fitness NMS and Bounded IoU Loss.
    <https://arxiv.org/abs/1711.00164>`_.

    Args:
        pred (torch.Tensor): Predicted bboxes.
        target (torch.Tensor): Target bboxes.
        beta (float): beta parameter in smoothl1.
        eps (float): eps to avoid NaN.
    """
    pred_ctrx = (pred[:, 0] + pred[:, 2]) * 0.5
    pred_ctry = (pred[:, 1] + pred[:, 3]) * 0.5
    pred_w = pred[:, 2] - pred[:, 0]
    pred_h = pred[:, 3] - pred[:, 1]
    with torch.no_grad():
        target_ctrx = (target[:, 0] + target[:, 2]) * 0.5
        target_ctry = (target[:, 1] + target[:, 3]) * 0.5
        target_w = target[:, 2] - target[:, 0]
        target_h = target[:, 3] - target[:, 1]

    dx = target_ctrx - pred_ctrx
    dy = target_ctry - pred_ctry

    loss_dx = 1 - torch.max(
        (target_w - 2 * dx.abs()) /
        (target_w + 2 * dx.abs() + eps), torch.zeros_like(dx))
    loss_dy = 1 - torch.max(
        (target_h - 2 * dy.abs()) /
        (target_h + 2 * dy.abs() + eps), torch.zeros_like(dy))
    loss_dw = 1 - torch.min(target_w / (pred_w + eps), pred_w /
                            (target_w + eps))
    loss_dh = 1 - torch.min(target_h / (pred_h + eps), pred_h /
                            (target_h + eps))
    loss_comb = torch.stack([loss_dx, loss_dy, loss_dw, loss_dh],
                            dim=-1).view(loss_dx.size(0), -1)

    loss = torch.where(loss_comb < beta, 0.5 * loss_comb * loss_comb / beta,
                       loss_comb - 0.5 * beta)
    return loss


@weighted_loss
def giou_loss(pred, target, eps=1e-7):
    r"""`Generalized Intersection over Union: A Metric and A Loss for Bounding
    Box Regression <https://arxiv.org/abs/1902.09630>`_.

    Args:
        pred (torch.Tensor): Predicted bboxes of format (x1, y1, x2, y2),
            shape (n, 4).
        target (torch.Tensor): Corresponding gt bboxes, shape (n, 4).
        eps (float): Eps to avoid log(0).

    Return:
        Tensor: Loss tensor.
    """
    gious = bbox_overlaps(pred, target, mode='giou', is_aligned=True, eps=eps)
    loss = 1 - gious
    return loss


@weighted_loss
def diou_loss(pred, target, eps=1e-7):
    r"""`Implementation of Distance-IoU Loss: Faster and Better
    Learning for Bounding Box Regression, https://arxiv.org/abs/1911.08287`_.

    Code is modified from https://github.com/Zzh-tju/DIoU.

    Args:
        pred (Tensor): Predicted bboxes of format (x1, y1, x2, y2),
            shape (n, 4).
        target (Tensor): Corresponding gt bboxes, shape (n, 4).
        eps (float): Eps to avoid log(0).
    Return:
        Tensor: Loss tensor.
    """
    # overlap
    lt = torch.max(pred[:, :2], target[:, :2])
    rb = torch.min(pred[:, 2:], target[:, 2:])
    wh = (rb - lt).clamp(min=0)
    overlap = wh[:, 0] * wh[:, 1]

    # union
    ap = (pred[:, 2] - pred[:, 0]) * (pred[:, 3] - pred[:, 1])
    ag = (target[:, 2] - target[:, 0]) * (target[:, 3] - target[:, 1])
    union = ap + ag - overlap + eps

    # IoU
    ious = overlap / union

    # enclose area
    enclose_x1y1 = torch.min(pred[:, :2], target[:, :2])
    enclose_x2y2 = torch.max(pred[:, 2:], target[:, 2:])
    enclose_wh = (enclose_x2y2 - enclose_x1y1).clamp(min=0)

    cw = enclose_wh[:, 0]
    ch = enclose_wh[:, 1]

    c2 = cw**2 + ch**2 + eps

    b1_x1, b1_y1 = pred[:, 0], pred[:, 1]
    b1_x2, b1_y2 = pred[:, 2], pred[:, 3]
    b2_x1, b2_y1 = target[:, 0], target[:, 1]
    b2_x2, b2_y2 = target[:, 2], target[:, 3]

    left = ((b2_x1 + b2_x2) - (b1_x1 + b1_x2))**2 / 4
    right = ((b2_y1 + b2_y2) - (b1_y1 + b1_y2))**2 / 4
    rho2 = left + right

    # DIoU
    dious = ious - rho2 / c2
    loss = 1 - dious
    return loss


@weighted_loss
def ciou_loss(pred, target, xyxy=True, eps=1e-7):
    r"""`Implementation of paper `Enhancing Geometric Factors into
    Model Learning and Inference for Object Detection and Instance
    Segmentation <https://arxiv.org/abs/2005.03572>`_.

    Code is modified from https://github.com/Zzh-tju/CIoU.

    Args:
        pred (Tensor): Predicted bboxes of format (x1, y1, x2, y2),
            shape (n, 4).
        target (Tensor): Corresponding gt bboxes, shape (n, 4).
        eps (float): Eps to avoid log(0).
    Return:
        Tensor: Loss tensor.
    """
    # if xyxy is false, make to xyxy
    if not xyxy:
        # cxcywh -> xyxy
        pred[:, :2] = pred[:, :2] - pred[:, 2:]/2
        pred[:, 2:] = pred[:, :2] + pred[:, 2:]

        target[:, :2] = target[:, :2] - target[:, 2:]/2
        target[:, 2:] = target[:, :2] + target[:, 2:]
    # overlap
    lt = torch.max(pred[:, :2], target[:, :2])
    rb = torch.min(pred[:, 2:], target[:, 2:])
    wh = (rb - lt).clamp(min=0)
    overlap = wh[:, 0] * wh[:, 1]

    # union
    ap = (pred[:, 2] - pred[:, 0]) * (pred[:, 3] - pred[:, 1])
    ag = (target[:, 2] - target[:, 0]) * (target[:, 3] - target[:, 1])
    union = ap + ag - overlap + eps

    # IoU
    ious = overlap / union

    # enclose area
    enclose_x1y1 = torch.min(pred[:, :2], target[:, :2])
    enclose_x2y2 = torch.max(pred[:, 2:], target[:, 2:])
    enclose_wh = (enclose_x2y2 - enclose_x1y1).clamp(min=0)

    cw = enclose_wh[:, 0]
    ch = enclose_wh[:, 1]

    c2 = cw**2 + ch**2 + eps

    b1_x1, b1_y1 = pred[:, 0], pred[:, 1]
    b1_x2, b1_y2 = pred[:, 2], pred[:, 3]
    b2_x1, b2_y1 = target[:, 0], target[:, 1]
    b2_x2, b2_y2 = target[:, 2], target[:, 3]

    w1, h1 = b1_x2 - b1_x1, b1_y2 - b1_y1 + eps
    w2, h2 = b2_x2 - b2_x1, b2_y2 - b2_y1 + eps

    left = ((b2_x1 + b2_x2) - (b1_x1 + b1_x2))**2 / 4
    right = ((b2_y1 + b2_y2) - (b1_y1 + b1_y2))**2 / 4
    rho2 = left + right

    factor = 4 / math.pi**2
    v = factor * torch.pow(torch.atan(w2 / h2) - torch.atan(w1 / h1), 2)

    # CIoU
    cious = ious - (rho2 / c2 + v**2 / (1 - ious + v))
    loss = 1 - cious
    return loss


def ciou(bboxes1, bboxes2, sum=True):
    """
    original ciou: https://github.com/Zzh-tju/CIoU/blob/master/layers/modules/multibox_loss.py
    bboxes1: xywh, bboxes2: xywh
    """
    bboxes1 = torch.sigmoid(bboxes1)
    bboxes2 = torch.sigmoid(bboxes2)
    rows = bboxes1.shape[0]
    cols = bboxes2.shape[0]
    cious = torch.zeros((rows, cols))
    if rows * cols == 0:
        return cious
    exchange = False
    if bboxes1.shape[0] > bboxes2.shape[0]:
        bboxes1, bboxes2 = bboxes2, bboxes1
        cious = torch.zeros((cols, rows))
        exchange = True
    w1 = torch.exp(bboxes1[:, 2])
    h1 = torch.exp(bboxes1[:, 3])
    w2 = torch.exp(bboxes2[:, 2])
    h2 = torch.exp(bboxes2[:, 3])
    area1 = w1 * h1
    area2 = w2 * h2
    center_x1 = bboxes1[:, 0]
    center_y1 = bboxes1[:, 1]
    center_x2 = bboxes2[:, 0]
    center_y2 = bboxes2[:, 1]

    inter_l = torch.max(center_x1 - w1 / 2, center_x2 - w2 / 2)
    inter_r = torch.min(center_x1 + w1 / 2, center_x2 + w2 / 2)
    inter_t = torch.max(center_y1 - h1 / 2, center_y2 - h2 / 2)
    inter_b = torch.min(center_y1 + h1 / 2, center_y2 + h2 / 2)
    inter_area = torch.clamp((inter_r - inter_l), min=0) * \
        torch.clamp((inter_b - inter_t), min=0)

    c_l = torch.min(center_x1 - w1 / 2, center_x2 - w2 / 2)
    c_r = torch.max(center_x1 + w1 / 2, center_x2 + w2 / 2)
    c_t = torch.min(center_y1 - h1 / 2, center_y2 - h2 / 2)
    c_b = torch.max(center_y1 + h1 / 2, center_y2 + h2 / 2)

    inter_diag = (center_x2 - center_x1)**2 + (center_y2 - center_y1)**2
    c_diag = torch.clamp((c_r - c_l), min=0)**2 + \
        torch.clamp((c_b - c_t), min=0)**2

    union = area1+area2-inter_area
    u = (inter_diag) / c_diag
    iou = inter_area / union
    v = (4 / (math.pi ** 2)) * \
        torch.pow((torch.atan(w2 / h2) - torch.atan(w1 / h1)), 2)
    with torch.no_grad():
        S = (iou > 0.5).float()
        alpha = S*v/(1-iou+v)
    cious = iou - u - alpha * v
    cious = torch.clamp(cious, min=-1.0, max=1.0)
    if exchange:
        cious = cious.T
    if sum:
        return torch.sum(1-cious)
    else:
        return 1-cious


def bbox_iou(box1, box2, x1y1x2y2=True, GIoU=False, DIoU=False, CIoU=False, eps=1e-7):
    # Returns the IoU of box1 to box2. box1 is 4, box2 is nx4
    box2 = box2.T

    # Get the coordinates of bounding boxes
    if x1y1x2y2:  # x1, y1, x2, y2 = box1
        b1_x1, b1_y1, b1_x2, b1_y2 = box1[0], box1[1], box1[2], box1[3]
        b2_x1, b2_y1, b2_x2, b2_y2 = box2[0], box2[1], box2[2], box2[3]
    else:  # transform from xywh to xyxy
        b1_x1, b1_x2 = box1[0] - box1[2] / 2, box1[0] + box1[2] / 2
        b1_y1, b1_y2 = box1[1] - box1[3] / 2, box1[1] + box1[3] / 2
        b2_x1, b2_x2 = box2[0] - box2[2] / 2, box2[0] + box2[2] / 2
        b2_y1, b2_y2 = box2[1] - box2[3] / 2, box2[1] + box2[3] / 2

    # Intersection area
    inter = (torch.min(b1_x2, b2_x2) - torch.max(b1_x1, b2_x1)).clamp(0) * \
            (torch.min(b1_y2, b2_y2) - torch.max(b1_y1, b2_y1)).clamp(0)

    # Union Area
    w1, h1 = b1_x2 - b1_x1, b1_y2 - b1_y1 + eps
    w2, h2 = b2_x2 - b2_x1, b2_y2 - b2_y1 + eps
    union = w1 * h1 + w2 * h2 - inter + eps

    iou = inter / union
    if GIoU or DIoU or CIoU:
        # convex (smallest enclosing box) width
        cw = torch.max(b1_x2, b2_x2) - torch.min(b1_x1, b2_x1)
        ch = torch.max(b1_y2, b2_y2) - torch.min(b1_y1, b2_y1)  # convex height
        if CIoU or DIoU:  # Distance or Complete IoU https://arxiv.org/abs/1911.08287v1
            c2 = cw ** 2 + ch ** 2 + eps  # convex diagonal squared
            rho2 = ((b2_x1 + b2_x2 - b1_x1 - b1_x2) ** 2 +
                    (b2_y1 + b2_y2 - b1_y1 - b1_y2) ** 2) / 4  # center distance squared
            if DIoU:
                return iou - rho2 / c2  # DIoU
            elif CIoU:  # https://github.com/Zzh-tju/DIoU-SSD-pytorch/blob/master/utils/box/box_utils.py#L47
                v = (4 / math.pi ** 2) * \
                    torch.pow(torch.atan(w2 / h2) - torch.atan(w1 / h1), 2)
                with torch.no_grad():
                    alpha = v / (v - iou + (1 + eps))
                return iou - (rho2 / c2 + v * alpha)  # CIoU
        else:  # GIoU https://arxiv.org/pdf/1902.09630.pdf
            c_area = cw * ch + eps  # convex area
            return iou - (c_area - union) / c_area  # GIoU
    else:
        return iou  # IoU


def bboxes_iou(boxes1, boxes2, x1y1x2y2=True, GIoU=False, DIoU=False, CIoU=False, eps=1e-7):
    """
    box1 is N,4 xyxy format
    box2 is N,4 xyxy format
    """
    n = boxes1.shape[0]
    res = torch.zeros([n, boxes2.shape[0]])
    for i in range(n):
        box1 = boxes1[i]
        a = bbox_iou(box1, boxes2, x1y1x2y2=x1y1x2y2,
                     GIoU=GIoU, DIoU=DIoU, CIoU=CIoU, eps=eps)
        res[i, :] = a
    return res



#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""

count how many certain files under dir

"""
import os
import glob
from alfred.utils.log import logger as logging



def count_file(d, f_type):
    assert os.path.exists(d), '{} not exist.'.format(d)
    # f_type can be jpg,png,pdf etc, connected by comma
    all_types = f_type.split(',')
    logging.info('count all file types: {} under: {}'.format(all_types, d))
    all_files = []
    for t in all_types:
        t = t.replace('.', '')
        one = glob.glob(os.path.join(d, '*.{}'.format(t)))
        one = [i for i in one if os.path.isfile(i)]
        logging.info('{} num: {}'.format(t, len(one)))
        all_files.extend(one)
    logging.info('file types: {}, total num: {}'.format(all_types, len(all_files)))



import time
import tabulate
import socket
import time
from threading import Thread
from queue import Queue
import cv2
import numpy as np
import os


def log(x):
    from datetime import datetime
    time_now = datetime.now().strftime("%m-%d-%H:%M:%S.%f ")
    print(time_now + x)


def myarray2string(array, separator=', ', fmt='%.3f', indent=8):
    assert len(array.shape) == 2, 'Only support MxN matrix, {}'.format(
        array.shape)
    blank = ' ' * indent
    res = ['[']
    for i in range(array.shape[0]):
        res.append(
            blank + '  ' + '[{}]'.format(separator.join([fmt % (d) for d in array[i]])))
        if i != array.shape[0] - 1:
            res[-1] += ', '
    res.append(blank + ']')
    return '\r\n'.join(res)


def write_common_results(dumpname=None, results=[], keys=[], fmt='%2.3f'):
    format_out = {'float_kind': lambda x: fmt % x}
    out_text = []
    out_text.append('[\n')
    for idata, data in enumerate(results):
        out_text.append('    {\n')
        output = {}
        output['id'] = data['id']
        for key in keys:
            if key not in data.keys():
                continue
            # BUG: This function will failed if the rows of the data[key] is too large
            # output[key] = np.array2string(data[key], max_line_width=1000, separator=', ', formatter=format_out)
            output[key] = myarray2string(data[key], separator=', ', fmt=fmt)
        for key in output.keys():
            out_text.append('        \"{}\": {}'.format(key, output[key]))
            if key != keys[-1]:
                out_text.append(',\n')
            else:
                out_text.append('\n')
        out_text.append('    }')
        if idata != len(results) - 1:
            out_text.append(',\n')
        else:
            out_text.append('\n')
    out_text.append(']\n')
    if dumpname is not None:
        os.makedirs(os.path.dirname(dumpname), exist_ok=True)
        with open(dumpname, 'w') as f:
            f.writelines(out_text)
    else:
        return ''.join(out_text)


def encode_detect(data):
    res = write_common_results(None, data, ['keypoints3d'])
    res = res.replace('\r', '').replace('\n', '').replace(' ', '')
    return res.encode('ascii')


def encode_smpl(data):
    res = write_common_results(
        None, data, ['poses', 'shapes', 'expression', 'Rh', 'Th'])
    res = res.replace('\r', '').replace('\n', '').replace(' ', '')
    return res.encode('ascii')


def encode_image(image):
    fourcc = [int(cv2.IMWRITE_JPEG_QUALITY), 90]
    result, img_encode = cv2.imencode('.jpg', image, fourcc)
    data = np.array(img_encode)
    stringData = data.tostring()
    return stringData


class BaseSocketClient:
    def __init__(self, host, port) -> None:
        if host == 'auto':
            host = socket.gethostname()
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((host, port))
        self.s = s

    def send(self, data):
        val = encode_detect(data)
        self.s.send(bytes('{}\n'.format(len(val)), 'ascii'))
        self.s.sendall(val)

    def send_smpl(self, data):
        val = encode_smpl(data)
        self.s.send(bytes('{}\n'.format(len(val)), 'ascii'))
        self.s.sendall(val)

    def close(self):
        self.s.close()


class BaseSocket:
    def __init__(self, host, port, debug=False) -> None:
        # 创建 socket 对象
        print('[Info] server start')
        serversocket = socket.socket(
            socket.AF_INET, socket.SOCK_STREAM)
        serversocket.bind((host, port))
        serversocket.listen(1)
        self.serversocket = serversocket
        self.queue = Queue()
        self.t = Thread(target=self.run)
        self.t.start()
        self.debug = debug
        self.disconnect = False

    @staticmethod
    def recvLine(sock):
        flag = True
        result = b''
        while not result.endswith(b'\n'):
            res = sock.recv(1)
            if not res:
                flag = False
                break
            result += res
        return flag, result.strip().decode('ascii')

    @staticmethod
    def recvAll(sock, l):
        l = int(l)
        result = b''
        while (len(result) < l):
            t = sock.recv(l - len(result))
            result += t
        return result.decode('ascii')

    def run(self):
        while True:
            clientsocket, addr = self.serversocket.accept()
            print("[Info] Connect: %s" % str(addr))
            self.disconnect = False
            while True:
                flag, l = self.recvLine(clientsocket)
                if not flag:
                    print("[Info] Disonnect: %s" % str(addr))
                    self.disconnect = True
                    break
                data = self.recvAll(clientsocket, l)
                if self.debug:
                    log('[Info] Recv data')
                self.queue.put(data)
            clientsocket.close()

    def update(self):
        time.sleep(1)
        while not self.queue.empty():
            log('update')
            data = self.queue.get()
            self.main(data)

    def main(self, datas):
        print(datas)

    def __del__(self):
        self.serversocket.close()
        self.t.join()


class Timer:
    records = {}
    tmp = None

    @classmethod
    def tic(cls):
        cls.tmp = time.time()

    @classmethod
    def toc(cls):
        res = (time.time() - cls.tmp) * 1000
        cls.tmp = None
        return res

    @classmethod
    def report(cls):
        header = ['', 'Time(ms)']
        contents = []
        for key, val in cls.records.items():
            contents.append(
                ['{:20s}'.format(key), '{:.2f}'.format(sum(val)/len(val))])
        print(tabulate.tabulate(contents, header, tablefmt='fancy_grid'))

    def __init__(self, name, silent=False):
        self.name = name
        self.silent = silent
        if name not in Timer.records.keys():
            Timer.records[name] = []

    def __enter__(self):
        self.start = time.time()

    def __exit__(self, exc_type, exc_value, exc_tb):
        end = time.time()
        Timer.records[self.name].append((end-self.start)*1000)
        if not self.silent:
            t = (end - self.start)*1000
            if t > 1000:
                print('-> [{:20s}]: {:5.1f}s'.format(self.name, t/1000))
            elif t > 1e3*60*60:
                print('-> [{:20s}]: {:5.1f}min'.format(self.name, t/1e3/60))
            else:
                print('-> [{:20s}]: {:5.1f}ms'.format(self.name,
                      (end-self.start)*1000))


'''
Color utils
'''


def generate_colorbar(N=20, cmap='jet'):
    bar = ((np.arange(N)/(N-1))*255).astype(np.uint8).reshape(-1, 1)
    colorbar = cv2.applyColorMap(bar, cv2.COLORMAP_JET).squeeze()
    if False:
        colorbar = np.clip(colorbar + 64, 0, 255)
    import random
    random.seed(666)
    index = [i for i in range(N)]
    random.shuffle(index)
    rgb = colorbar[index, :]
    rgb = rgb.tolist()
    return rgb


colors_bar_rgb = generate_colorbar(cmap='hsv')

colors_table = {
    'b': [0.65098039, 0.74117647, 0.85882353],
    '_pink': [.9, .7, .7],
    '_mint': [166/255.,  229/255.,  204/255.],
    '_mint2': [202/255.,  229/255.,  223/255.],
    '_green': [153/255.,  216/255.,  201/255.],
    '_green2': [171/255.,  221/255.,  164/255.],
    'r': [251/255.,  128/255.,  114/255.],
    '_orange': [253/255.,  174/255.,  97/255.],
    'y': [250/255.,  230/255.,  154/255.],
    '_r': [255/255, 0, 0],
    'g': [0, 255/255, 0],
    '_b': [0, 0, 255/255],
    'k': [0, 0, 0],
    '_y': [255/255, 255/255, 0],
    'purple': [128/255, 0, 128/255],
    'smap_b': [51/255, 153/255, 255/255],
    'smap_r': [255/255, 51/255, 153/255],
    'smap_b': [51/255, 255/255, 153/255],
}


def get_rgb(index):
    if isinstance(index, int):
        if index == -1:
            return (255, 255, 255)
        if index < -1:
            return (0, 0, 0)
        col = colors_bar_rgb[index % len(colors_bar_rgb)]
    else:
        col = colors_table.get(index, (1, 0, 0))
        col = tuple([int(c*255) for c in col[::-1]])
    return col


def get_rgb_01(index):
    col = get_rgb(index)
    return [i*1./255 for i in col[:3]]


class BaseCrit:
    def __init__(self, min_conf, min_joints=3) -> None:
        self.min_conf = min_conf
        self.min_joints = min_joints
        self.name = self.__class__.__name__

    def __call__(self, keypoints3d, **kwargs):
        # keypoints3d: (N, 4)
        conf = keypoints3d[..., -1]
        conf[conf < self.min_conf] = 0
        idx = keypoints3d[..., -1] > self.min_conf
        return len(idx) > self.min_joints


class CritRange(BaseCrit):
    def __init__(self, minr, maxr, rate_inlier, min_conf) -> None:
        super().__init__(min_conf)
        self.min = minr
        self.max = maxr
        self.rate = rate_inlier

    def __call__(self, keypoints3d, **kwargs):
        idx = keypoints3d[..., -1] > self.min_conf
        k3d = keypoints3d[idx, :3]
        crit = (k3d[:, 0] > self.min[0]) & (k3d[:, 0] < self.max[0]) &\
            (k3d[:, 1] > self.min[1]) & (k3d[:, 1] < self.max[1]) &\
            (k3d[:, 2] > self.min[2]) & (k3d[:, 2] < self.max[2])
        self.log = '{}: {}'.format(self.name, k3d)
        return crit.sum()/crit.shape[0] > self.rate



#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""

this tool helps viewing VOC format data


"""
import os
import sys
import cv2
import xml.etree.ElementTree as ET
from glob import glob
import os
import sys
import cv2
from alfred.utils.log import logger as logging


def get_and_check(root, name, length):
    vars = root.findall(name)
    if len(vars) == 0:
        raise NotImplementedError('Can not find %s in %s.' % (name, root.tag))
    if length > 0 and len(vars) != length:
        raise NotImplementedError('The size of %s is supposed to be %d, but is %d.' % (name, length, len(vars)))
    if length == 1:
        vars = vars[0]
    return vars


def get(root, name):
    vars = root.findall(name)
    return vars


def vis_voc(img_root, label_root, label_major=True):
    logging.info('img root: {}, label root: {}'.format(img_root, label_root))
    # auto detection .jpg or .png images
    if label_major:
        logging.info('label major will using xmls to found images... it might cause no image found')
        xml_files = glob(os.path.join(label_root, '*.xml'))
        for xml in xml_files:
            if os.path.exists(xml):
                img_f = os.path.join(img_root, os.path.basename(xml).split('.')[0] + '.jpg')
                if not os.path.exists(img_f):
                    logging.info('switch to png format')
                    img_f = os.path.join(img_root, os.path.basename(xml).split('.')[0] + '.png')
                img = cv2.imread(img_f)
                if os.path.exists(img_f):
                    tree = ET.parse(xml)
                    root = tree.getroot()
                    for obj in get(root, 'object'):
                        category = get_and_check(obj, 'name', 1).text
                        bndbox = get_and_check(obj, 'bndbox', 1)
                        xmin = int(float(get_and_check(bndbox, 'xmin', 1).text))
                        ymin = int(float(get_and_check(bndbox, 'ymin', 1).text))
                        xmax = int(float(get_and_check(bndbox, 'xmax', 1).text))
                        ymax = int(float(get_and_check(bndbox, 'ymax', 1).text))

                        cv2.putText(img, category, (xmin, ymin), cv2.FONT_HERSHEY_COMPLEX, 0.7, (255, 255, 255))
                        cv2.rectangle(img, (xmin, ymin), (xmax, ymax), (0, 255, 0), 2, 1)
                    cv2.imshow('voc check', img)
                    cv2.waitKey(0)
                else:
                    logging.warning('xxxx image: {} for label: {} not found.'.format(img_f, xml))
    else:
        img_files = glob(os.path.join(img_root, '*.[jp][pn]g'))
        for img_f in img_files:
            if os.path.exists(img_f):
                img = cv2.imread(img_f)
                label_path = os.path.join(label_root, os.path.basename(img_f).split('.')[0] + '.xml')
                if os.path.exists(label_path):
                    #
                    tree = ET.parse(label_path)
                    root = tree.getroot()
                    for obj in get(root, 'object'):
                        category = get_and_check(obj, 'name', 1).text
                        bndbox = get_and_check(obj, 'bndbox', 1)
                        xmin = int(float(get_and_check(bndbox, 'xmin', 1).text))
                        ymin = int(float(get_and_check(bndbox, 'ymin', 1).text))
                        xmax = int(float(get_and_check(bndbox, 'xmax', 1).text))
                        ymax = int(float(get_and_check(bndbox, 'ymax', 1).text))

                        cv2.putText(img, category, (xmin, ymin), cv2.FONT_HERSHEY_COMPLEX, 0.7, (255, 255, 255))
                        cv2.rectangle(img, (xmin, ymin), (xmax, ymax), (0, 255, 0), 2, 1)
                    cv2.imshow('voc check', img)
                    cv2.waitKey(0)
                else:
                    logging.warning('xxxx image: {} according label: {} not found.'.format(img_f, label_path))


if __name__ == "__main__":
    vis_voc(sys.argv[1], sys.argv[2])



import platform
import os


def mkdir(path):
    e = os.path.exists(path)
    if not e:
        os.makedirs(path)
        return True
    else:
        return False


def mkfile(filePath):
    pipfile = "[global]\ntrusted-host=mirrors.aliyun.com\nindex-url=http://mirrors.aliyun.com/pypi/simple/"
    if os.path.exists(filePath):
        if str(input("File exist!Cover?(Y/N))")).upper() == 'N':
            print("Not Cover.")
            return
    with open(filePath, 'w') as fp:
        fp.write(pipfile)
    print("Write finish.")


def change_pypi_source():
    systype = platform.system()
    print("System type: " + systype)
    if systype == "Windows":
        path = os.path.join(os.getenv('HOMEPATH'), 'pip')
        mkdir(path)
        mkfile(os.path.join(path, 'pip.ini'))
    elif systype == "Linux" or systype == "Darwin":
        path = os.path.join(os.path.expandvars('$HOME'), ".pip")
        mkdir(path)
        mkfile(os.path.join(path, 'pip.conf'))
    else:
        print("System type: " + systype + " Not Support!")



#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
showing 3d point cloud using open3d
"""
import numpy as np
import cv2
import matplotlib.pyplot as plt

from alfred.fusion.common import compute_3d_box_lidar_coords
try:
    from open3d import *
    import open3d as o3d
except ImportError:
    print('importing 3d_vis in alfred-py need open3d installed.')
    exit(0)


def draw_pointclouds_boxes_o3d(pointcloud, boxes_3d, line_color=[0, 1, 0]):
    """
    draw boxes3d on pointcloud, a typical boxes_3d format:
    [[4.481686, 5.147319, -1.0229858, 1.5728549, 3.646751, 1.5121397, 1.5486346],
       [-2.5172017, 5.0262384, -1.0679419, 1.6241353,
           4.0445814, 1.4938312, 1.620804]]
    which is:
    [[x,y,z,w,h,l,roy]]
    """
    geometries = []

    pcs = np.array(pointcloud[:, :3])
    pcobj = o3d.geometry.PointCloud()
    pcobj.points = o3d.utility.Vector3dVector(pcs)
    geometries.append(pcobj)

    # append boxes to geometries
    for p in boxes_3d:
        xyz = np.array([p[: 3]])
        hwl = np.array([p[3: 6]])
        r_y = [p[6]]
        pts3d = compute_3d_box_lidar_coords(
            xyz, hwl, angles=r_y, origin=(0.5, 0.5, 0.5), axis=2)

        lines = [[0, 1], [1, 2], [2, 3], [3, 0],
                 [4, 5], [5, 6], [6, 7], [7, 4],
                 [0, 4], [1, 5], [2, 6], [3, 7]]
        colors = [line_color for i in range(len(lines))]
        line_set = o3d.geometry.LineSet()
        line_set.points = o3d.utility.Vector3dVector(pts3d[0])
        line_set.lines = o3d.utility.Vector2iVector(lines)
        line_set.colors = o3d.utility.Vector3dVector(colors)
        geometries.append(line_set)

    def capture_depth(vis):
        depth = vis.capture_depth_float_buffer()
        plt.imshow(np.asarray(depth))
        plt.show()
        return False

    def capture_image(vis):
        image = vis.capture_screen_float_buffer()
        plt.imshow(np.asarray(image))
        plt.show()
        return False

    vis = o3d.visualization.Visualizer()
    vis.create_window()
    for g in geometries:
        vis.add_geometry(g)
    opt = vis.get_render_option()
    opt.background_color = np.asarray([255, 255, 255])
    opt.point_size = 2
    opt.line_width = 3
    # opt.show_coordinate_frame = True
    vis.run()
    vis.destroy_window()


def draw_pcs_open3d(geometries):
    """
    drawing the points using open3d
    it can draw points and linesets
    ```
    point_cloud = PointCloud()
    point_cloud.points = Vector3dVector(pcs)


    points = [[0,0,0],[1,0,0],[0,1,0],[1,1,0],
                [0,0,1],[1,0,1],[0,1,1],[1,1,1]]
    lines = [[0,1],[0,2],[1,3],[2,3],
                [4,5],[4,6],[5,7],[6,7],
                [0,4],[1,5],[2,6],[3,7]]
    colors = [[1, 0, 0] for i in range(len(lines))]
    line_set = LineSet()
    line_set.points = Vector3dVector(points)
    line_set.lines = Vector2iVector(lines)
    line_set.colors = Vector3dVector(colors)
    draw_pcs_open3d([point_cloud, line_set])
    ```
    """
    def capture_depth(vis):
        depth = vis.capture_depth_float_buffer()
        plt.imshow(np.asarray(depth))
        plt.show()
        return False

    def capture_image(vis):
        image = vis.capture_screen_float_buffer()
        plt.imshow(np.asarray(image))
        plt.show()
        return False
    vis = o3d.visualization.Visualizer()
    vis.create_window()
    for g in geometries:
        vis.add_geometry(g)
    opt = vis.get_render_option()
    opt.background_color = np.asarray([255, 255, 255])
    opt.point_size = 2
    opt.show_coordinate_frame = True
    vis.run()
    vis.destroy_window()



import xml.etree.ElementTree as ET
import pickle
import os
from os import listdir, getcwd
from os.path import join
import sys
import glob



def convert(size, box):
    dw = 1./(size[0])
    dh = 1./(size[1])
    x = (box[0] + box[1])/2.0 - 1
    y = (box[2] + box[3])/2.0 - 1
    w = box[1] - box[0]
    h = box[3] - box[2]
    x = x*dw
    w = w*dw
    y = y*dh
    h = h*dh
    return (x, y, w, h)


def convert_annotation(xml_f, target_dir, classes_names):
    f_name = os.path.basename(xml_f).split('.')[0] + '.txt'
    out_file = open(os.path.join(target_dir, f_name), 'w')

    tree = ET.parse(xml_f)
    root = tree.getroot()
    size = root.find('size')
    w = int(size.find('width').text)
    h = int(size.find('height').text)

    for obj in root.iter('object'):
        difficult = obj.find('difficult').text
        cls = obj.find('name').text
        if cls not in classes_names or int(difficult) == 1:
            continue
        cls_id = classes_names.index(cls)
        xmlbox = obj.find('bndbox')
        b = (float(xmlbox.find('xmin').text), float(xmlbox.find('xmax').text), float(
            xmlbox.find('ymin').text), float(xmlbox.find('ymax').text))
        bb = convert((w, h), b)
        out_file.write(str(cls_id) + " " +
                       " ".join([str(a) for a in bb]) + '\n')



def voc2yolo(img_dir, xml_dir, class_txt):
    classes_names = None
    if class_txt:
        classes_names = [i.strip() for i in open(class_txt, 'r').readlines()]

    labels_target = os.path.join(os.path.dirname(xml_dir.rstrip('/')), 'yolo_converted_from_voc')
    print('labels dir to save: {}'.format(labels_target))
    if not os.path.exists(labels_target):
        os.makedirs(labels_target)

    xmls = glob.glob(os.path.join(xml_dir, '*.xml'))
    for xml in xmls:
        convert_annotation(xml, labels_target, classes_names)
    print('Done!')
    print('class name order used is: ', classes_names)


#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
draw detection result base on various format

after detection

also include draw 3d box on image
"""
import numpy as np
import cv2
import os
import random
from .common import create_unique_color_uchar, get_unique_color_by_id2, colors
from .common import draw_rect_with_style
import warnings
from collections import Counter, OrderedDict
from .common import put_txt_with_newline
from .get_dataset_label_map import coco_label_map_list


def _draw_round_dot_border(img, pt1, pt2, color, thickness, r=2, d=5):
    x1, y1 = pt1
    x2, y2 = pt2

    # Top left
    cv2.line(img, (x1 + r, y1), (x1 + r + d, y1), color, thickness)
    cv2.line(img, (x1, y1 + r), (x1, y1 + r + d), color, thickness)
    cv2.ellipse(img, (x1 + r, y1 + r), (r, r), 180, 0, 90, color, thickness)

    # Top right
    cv2.line(img, (x2 - r, y1), (x2 - r - d, y1), color, thickness)
    cv2.line(img, (x2, y1 + r), (x2, y1 + r + d), color, thickness)
    cv2.ellipse(img, (x2 - r, y1 + r), (r, r), 270, 0, 90, color, thickness)

    # Bottom left
    cv2.line(img, (x1 + r, y2), (x1 + r + d, y2), color, thickness)
    cv2.line(img, (x1, y2 - r), (x1, y2 - r - d), color, thickness)
    cv2.ellipse(img, (x1 + r, y2 - r), (r, r), 90, 0, 90, color, thickness)

    # Bottom right
    cv2.line(img, (x2 - r, y2), (x2 - r - d, y2), color, thickness)
    cv2.line(img, (x2, y2 - r), (x2, y2 - r - d), color, thickness)
    cv2.ellipse(img, (x2 - r, y2 - r), (r, r), 0, 0, 90, color, thickness)
    return img


def draw_one_bbox(image, box, unique_color, thickness):
    x1 = int(box[0])
    y1 = int(box[1])
    x2 = int(box[2])
    y2 = int(box[3])
    cv2.rectangle(image, (x1, y1), (x2, y2), unique_color, thickness)
    return image


# ==================== Below are deprecation API =================
def draw_box_without_score(img, boxes, classes=None, is_show=False):
    """
    Draw boxes on image, the box mostly are annotations, not the model predict box
    """
    warnings.warn(
        'this method is deprecated, using visiualize_det_cv2 instead', DeprecationWarning)
    assert isinstance(boxes,
                      np.ndarray), 'boxes must nump array, with shape of (None, 5)\nevery element contains (x1,y1,x2,y2, label)'
    if classes:
        pass
    else:
        height = img.shape[0]
        width = img.shape[1]

        font = cv2.QT_FONT_NORMAL
        font_scale = 0.4
        font_thickness = 1
        line_thickness = 1

        all_cls = []
        for i in range(boxes.shape[0]):
            cls = boxes[i, -1]
            all_cls.append(cls)
            all_cls = set(all_cls)
            unique_color = create_unique_color_uchar(all_cls.index(cls))

            y1 = int(boxes[i, 2])
            x1 = int(boxes[i, 3])
            y2 = int(boxes[i, 4])
            x2 = int(boxes[i, 5])

            cv2.rectangle(img, (x1, y1), (x2, y2),
                          unique_color, line_thickness)

            text_label = '{}'.format(cls)
            (ret_val, base_line) = cv2.getTextSize(
                text_label, font, font_scale, font_thickness)
            text_org = (x1, y1 - 0)

            cv2.rectangle(img, (text_org[0] - 5, text_org[1] + base_line + 2),
                          (text_org[0] + ret_val[0] + 5, text_org[1] - ret_val[1] - 2), unique_color, line_thickness)
            # this rectangle for fill text rect
            cv2.rectangle(img, (text_org[0] - 5, text_org[1] + base_line + 2),
                          (text_org[0] + ret_val[0] + 4,
                           text_org[1] - ret_val[1] - 2),
                          unique_color, -1)
            cv2.putText(img, text_label, text_org, font,
                        font_scale, (255, 255, 255), font_thickness)
        if is_show:
            cv2.imshow('image', img)
            cv2.waitKey(0)
        return img


def visualize_det_cv2(img, detections, classes=None, thresh=0.6, is_show=False, background_id=-1, mode='xyxy'):
    """
    visualize detection on image using cv2, this is the standard way to visualize detections

    new add mode option
    mode can be one of 'xyxy' and 'xywh', 'xyxy' as default

    :param img:
    :param detections: ssd detections, numpy.array([[id, score, x1, y1, x2, y2]...])
            each row is one object
    :param classes:
    :param thresh:
    :param is_show:
    :param background_id: -1
    :param mode:
    :return:
    """
    assert classes, 'from visualize_det_cv2, classes must be provided, each class in a list with' \
                    'certain order.'
    assert isinstance(
        img, np.ndarray), 'from visualize_det_cv2, img must be a numpy array object.'

    font = cv2.FONT_HERSHEY_SIMPLEX
    font_scale = 0.36

    font_thickness = 1
    line_thickness = 1

    for i in range(detections.shape[0]):
        cls_id = int(detections[i, 0])
        if cls_id != background_id:
            score = detections[i, 1]
            if score > thresh:
                unique_color = create_unique_color_uchar(cls_id)
                x1, y1, x2, y2 = 0, 0, 0, 0
                if mode == 'xyxy':
                    x1 = int(detections[i, 2])
                    y1 = int(detections[i, 3])
                    x2 = int(detections[i, 4])
                    y2 = int(detections[i, 5])
                else:
                    x1 = int(detections[i, 2])
                    y1 = int(detections[i, 3])
                    x2 = x1 + int(detections[i, 4])
                    y2 = y1 + int(detections[i, 5])

                cv2.rectangle(img, (x1, y1), (x2, y2),
                              unique_color, line_thickness, cv2.LINE_AA)
                text_label = '{} {:.2f}'.format(classes[cls_id], score)
                (ret_val, _) = cv2.getTextSize(
                    text_label, font, font_scale, font_thickness)
                txt_bottom_left = (x1+4, y1-4)
                cv2.rectangle(img, (txt_bottom_left[0]-4, txt_bottom_left[1] - ret_val[1]-2),
                              (txt_bottom_left[0] + ret_val[0] +
                               2, txt_bottom_left[1]+4),
                              (0, 0, 0), -1)
                cv2.putText(img, text_label, txt_bottom_left, font,
                            font_scale, (237, 237, 237), font_thickness, cv2.LINE_AA)
    if is_show:
        cv2.imshow('image', img)
        cv2.waitKey(0)
    return img


def visualize_det_cv2_style0(img, detections, classes=None, cls_colors=None, thresh=0.3, suit_color=True, blend=False,
                             is_show=False, background_id=-1, mode='xyxy', line_thickness=1,
                             font_scale=0.48, counter_on=False, text_bk=False, counter_pos=(30, 150)):
    """
    visualize detection on image using cv2, this is the standard way to visualize detections

    new add mode option
    mode can be one of 'xyxy' and 'xywh', 'xyxy' as default

    :param img:
    :param detections: ssd detections, numpy.array([[id, score, x1, y1, x2, y2]...])
            each row is one object
    :param classes:
    :param cls_colors:
    :param thresh:
    :param is_show:
    :param background_id: -1
    :param mode:
    :return:
    """
    assert isinstance(
        img, np.ndarray), 'from visualize_det_cv2, img must be a numpy array object.'
    if cls_colors and classes:
        assert len(cls_colors) == len(
            classes), 'cls_colors must be same with classes length if you specific cls_colors.'

    font = cv2.FONT_HERSHEY_SIMPLEX
    font_thickness = 1

    if blend:
        raw_img = img.copy()
        img = np.zeros_like(img)

    cls_counter = []
    for i in range(detections.shape[0]):
        cls_id = int(detections[i, 0])
        if cls_id != background_id:
            score = detections[i, 1]
            if score > thresh:
                if cls_colors:
                    unique_color = cls_colors[cls_id]
                else:
                    unique_color = get_unique_color_by_id2(cls_id)
                x1, y1, x2, y2 = 0, 0, 0, 0
                if mode == 'xyxy':
                    x1 = int(detections[i, 2])
                    y1 = int(detections[i, 3])
                    x2 = int(detections[i, 4])
                    y2 = int(detections[i, 5])
                else:
                    x1 = int(detections[i, 2])
                    y1 = int(detections[i, 3])
                    x2 = x1 + int(detections[i, 4])
                    y2 = y1 + int(detections[i, 5])

                cv2.rectangle(img, (x1, y1), (x2, y2),
                              unique_color, line_thickness, cv2.LINE_AA)
                if classes:
                    text_label = '{} {:.1f}%'.format(
                        classes[cls_id], score*100)
                    if counter_on:
                        cls_counter.append(classes[cls_id])
                else:
                    text_label = '{} {:.1f}%'.format(cls_id, score*100)

                ((txt_w, txt_h), _) = cv2.getTextSize(
                    text_label, font, font_scale, 1)
                # Place text background.
                back_tl = x1, y1 - int(1.5*txt_h) - 4
                back_br = x1 + 8 + txt_w, y1
                if text_bk:
                    if suit_color:
                        cv2.rectangle(img, back_tl, back_br, unique_color, -1)
                    else:
                        cv2.rectangle(img, back_tl, back_br, (0, 0, 0), -1)
                    # 2 pixel offset from left
                    txt_tl = x1+4, y1 - int(0.5 * txt_h)
                else:
                    txt_tl = x1-txt_w//2 + (x2-x1)//2, y1 - int(0.5 * txt_h)
                if text_bk:
                    cv2.putText(img, text_label, txt_tl, font, font_scale,
                                (0, 0, 0), font_thickness, cv2.LINE_AA)
                else:
                    cv2.putText(img, text_label, txt_tl, font, font_scale,
                                unique_color, font_thickness, cv2.LINE_AA)
    if counter_on:
        cc = Counter(cls_counter)
        cc = OrderedDict(sorted(cc.items()))
        # drw counter result on image
        txt = ''
        for k, v in cc.items():
            txt += '{}: {}\n'.format(k, v)
        put_txt_with_newline(img, txt, counter_pos, font,
                             1.9, (0, 255, 0), 2, cv2.LINE_AA)

    if blend:
        img = cv2.addWeighted(raw_img, 1.0, img, 0.9, 0.6)

    if is_show:
        cv2.imshow('image', img)
        cv2.waitKey(0)
    return img


def visualize_det_cv2_fancy(img, detections, classes=None, thresh=0.2, is_show=False, background_id=-1, mode='xyxy', r=4, d=6):
    """
    visualize detections with a more fancy way

    new add mode option
    mode can be one of 'xyxy' and 'xywh', 'xyxy' as default

    :param img:
    :param detections: ssd detections, numpy.array([[id, score, x1, y1, x2, y2]...])
            each row is one object
    :param classes:
    :param thresh:
    :param is_show:
    :param background_id: -1
    :param mode:
    :return:
    """
    assert isinstance(
        img, np.ndarray), 'from visualize_det_cv2, img must be a numpy array object.'

    height = img.shape[0]
    width = img.shape[1]

    font = cv2.FONT_HERSHEY_SIMPLEX
    font_scale = 0.35
    font_thickness = 1
    line_thickness = 2

    for i in range(detections.shape[0]):
        cls_id = int(detections[i, 0])
        if cls_id != background_id:
            score = detections[i, 1]
            if score > thresh:
                unique_color = create_unique_color_uchar(cls_id)
                x1, y1, x2, y2 = 0, 0, 0, 0
                if mode == 'xyxy':
                    x1 = int(detections[i, 2])
                    y1 = int(detections[i, 3])
                    x2 = int(detections[i, 4])
                    y2 = int(detections[i, 5])
                else:
                    x1 = int(detections[i, 2])
                    y1 = int(detections[i, 3])
                    x2 = x1 + int(detections[i, 4])
                    y2 = y1 + int(detections[i, 5])

                if classes:
                    text_label = '{} {:.1f}%'.format(
                        classes[cls_id], score*100)
                    if counter_on:
                        cls_counter.append(classes[cls_id])
                else:
                    text_label = '{} {:.1f}%'.format(cls_id, score*100)

                _draw_round_dot_border(
                    img, (x1, y1), (x2, y2), unique_color, line_thickness, r, d)
                (txt_size, line_h) = cv2.getTextSize(
                    text_label, font, font_scale, font_thickness)
                txt_org = (int((x1+x2)/2 - txt_size[0]/2), int(y1+line_h+2))
                cv2.putText(img, text_label, txt_org, font, font_scale,
                            (255, 255, 255), font_thickness, lineType=cv2.LINE_AA)
    if is_show:
        cv2.imshow('image', img)
        cv2.waitKey(0)
    return img


def visualize_det_cv2_part(img, scores, cls_ids, boxes, class_names=None, thresh=0.2,
                           is_show=False, random=False, background_id=-1, mode='xyxy', style='none',
                           force_color=None, line_thickness=2, font_scale=0.2, show_time=0, wait_t=0, transparent=False):
    """
    visualize detection on image using cv2, this is the standard way to visualize detections

    new add mode option
    mode can be one of 'xyxy' and 'xywh', 'xyxy' as default

    :param img:
    :param detections: ssd detections, numpy.array([[id, score, x1, y1, x2, y2]...])
            each row is one object
    :param classes:
    :param thresh:
    :param is_show:
    :param background_id: -1
    :param mode:
    :return:
    """
    assert isinstance(
        img, np.ndarray), 'from visualize_det_cv2, img must be a numpy array object.'
    if force_color:
        assert isinstance(force_color, list) or isinstance(
            force_color, np.ndarray), 'force_color must be list or numpy array'

    font = cv2.FONT_HERSHEY_SIMPLEX
    font_thickness = 1
    line_thickness = line_thickness

    if transparent:
        img_mask = np.zeros_like(img)

    n_boxes = 0
    if isinstance(boxes, np.ndarray):
        n_boxes = boxes.shape[0]
    elif isinstance(boxes, list):
        n_boxes = len(boxes)
    else:
        print('boxes with unsupported type, boxes must be ndarray or list.')

    if class_names is None or len(class_names) == 0:
        # not using background
        class_names = coco_label_map_list[1:]

    for i in range(n_boxes):
        cls_id = int(cls_ids[i])
        if cls_id != background_id:
            if scores is not None:
                if scores[i] > thresh:
                    if force_color:
                        if random:
                            unique_color = force_color[np.random.randint(100)]
                        else:
                            if cls_id > len(force_color)-1:
                                unique_color = force_color[min(
                                    cls_id, len(force_color)-1)]
                            else:
                                unique_color = force_color[cls_id]
                    else:
                        unique_color = colors(cls_id, True)
                        # unique_color = colors[cls_id]
                    x1, y1, x2, y2 = 0, 0, 0, 0
                    if mode == 'xyxy':
                        x1 = int(boxes[i, 0])
                        y1 = int(boxes[i, 1])
                        x2 = int(boxes[i, 2])
                        y2 = int(boxes[i, 3])
                    else:
                        x1 = int(boxes[i, 0])
                        y1 = int(boxes[i, 1])
                        x2 = x1 + int(boxes[i, 2])
                        y2 = y1 + int(boxes[i, 3])

                    if style in ['dashed', 'dotted']:
                        draw_rect_with_style(
                            img, (x1, y1), (x2, y2), unique_color, line_thickness, style=style)
                    else:
                        if transparent:
                            cv2.rectangle(img_mask, (x1, y1), (x2, y2),
                                          unique_color, line_thickness, cv2.LINE_AA)
                        else:
                            cv2.rectangle(img, (x1, y1), (x2, y2),
                                          unique_color, line_thickness, cv2.LINE_AA)

                    if class_names:
                        if cls_id > len(class_names) - 1:
                            n = class_names[min(cls_id, len(force_color)-1)]
                        else:
                            n = class_names[cls_id]
                        text_label = '{} {:.1f}%'.format(n, scores[i]*100)
                    else:
                        text_label = '{} {:.1f}%'.format(cls_id, scores[i]*100)

                    (ret_val, _) = cv2.getTextSize(
                        text_label, font, font_scale, font_thickness)
                    # txt_bottom_left = (x1+4, y1-4)
                    txt_bottom_left = (x1, y1-6) # make text a little higher
                    if transparent:
                        cv2.rectangle(img_mask, (txt_bottom_left[0], txt_bottom_left[1] - ret_val[1]-6),
                                      (txt_bottom_left[0] + ret_val[0]
                                       , txt_bottom_left[1]+6),
                                      unique_color, -1, cv2.LINE_AA)
                    else:
                        cv2.rectangle(img, (txt_bottom_left[0], txt_bottom_left[1] - ret_val[1]-6),
                                      (txt_bottom_left[0] + ret_val[0]
                                       , txt_bottom_left[1]+6),
                                      unique_color, -1, cv2.LINE_AA)
                    cv2.putText(img, text_label, txt_bottom_left, font,
                                font_scale, (237, 237, 237), font_thickness, cv2.LINE_AA)
            else:
                if force_color:
                    if random:
                        unique_color = force_color[np.random.randint(100)]
                    else:
                        unique_color = force_color[cls_id]
                else:
                    unique_color = colors(cls_id, True)
                    # unique_color = colors[cls_id]
                x1, y1, x2, y2 = 0, 0, 0, 0
                if mode == 'xyxy':
                    x1 = int(boxes[i, 0])
                    y1 = int(boxes[i, 1])
                    x2 = int(boxes[i, 2])
                    y2 = int(boxes[i, 3])
                else:
                    x1 = int(boxes[i, 0])
                    y1 = int(boxes[i, 1])
                    x2 = x1 + int(boxes[i, 2])
                    y2 = y1 + int(boxes[i, 3])

                if style in ['dashed', 'dotted']:
                    draw_rect_with_style(
                        img, (x1, y1), (x2, y2), unique_color, line_thickness, style=style)
                else:
                    if transparent:
                        cv2.rectangle(img_mask, (x1, y1), (x2, y2),
                                      unique_color, line_thickness, cv2.LINE_AA)
                    else:
                        cv2.rectangle(img, (x1, y1), (x2, y2),
                                      unique_color, line_thickness, cv2.LINE_AA)

                if class_names:
                    text_label = '{}'.format(class_names[cls_id])
                else:
                    text_label = '{}'.format(cls_id)

                (ret_val, _) = cv2.getTextSize(
                    text_label, font, font_scale, font_thickness)
                txt_bottom_left = (x1+4, y1-4)
                if transparent:
                    cv2.rectangle(img_mask, (txt_bottom_left[0]-4, txt_bottom_left[1] - ret_val[1]-2),
                                  (txt_bottom_left[0] + ret_val[0] +
                                   2, txt_bottom_left[1]+4),
                                  unique_color, -1, cv2.LINE_AA)
                else:
                    cv2.rectangle(img, (txt_bottom_left[0]-4, txt_bottom_left[1] - ret_val[1]-2),
                                  (txt_bottom_left[0] + ret_val[0] +
                                   2, txt_bottom_left[1]+4),
                                  unique_color, -1, cv2.LINE_AA)
                cv2.putText(img, text_label, txt_bottom_left, font,
                            font_scale, (237, 237, 237), font_thickness, cv2.LINE_AA)
    if show_time > 0:
        cv2.putText(img, f'{round(1/show_time, 1)}FPS {show_time*1000}ms', (46, 60), font,
                    font_scale+0.6, (125, 209, 46), 2, cv2.LINE_AA)

    if transparent:
        img = cv2.addWeighted(img_mask, 0.99, img, 0.7, 0.9)

    if is_show:
        cv2.imshow('image', img)
        cv2.waitKey(wait_t)
    return img


def visualize_det_mask_cv2(img, detections, masks, classes=None, is_show=False, background_id=-1, is_video=False):
    """
    this method using for display detections and masks on image
    :param img:
    :param detections: numpy.array([[id, score, x1, y1, x2, y2]...])
            each row is one object. contains id and score in the first 2 rows
    :param masks: numpy.array([[mask_width, mask_height], ...], every element is an
    one chanel mask of on object
    :param classes: classes names in a list with certain order
    :param is_show: to show if it is video
    :param background_id
    :param is_video
    :return:
    """
    assert isinstance(img, np.ndarray) and isinstance(detections, np.ndarray) and isinstance(masks, np.ndarray), \
        'images and detections and masks must be numpy array'
    assert detections.shape[0] == masks.shape[-1], 'detections nums and masks nums are not equal'
    assert is_show != is_video, 'you can not set is_show and is_video at the same time.'
    # draw detections first
    img = visualize_det_cv2(img, detections, classes=classes, is_show=False)

    masked_image = img
    print('masked image shape: ', masked_image.shape)
    num_instances = detections.shape[0]
    for i in range(num_instances):
        cls_id = int(detections[i, 0])
        if cls_id != background_id:
            unique_color = create_unique_color_uchar(cls_id)
            mask = masks[:, :, i]
            masked_image = _apply_mask2(masked_image, mask, unique_color)
    # masked_image = masked_image.astype(int)
    if is_video:
        cv2.imshow('image', masked_image)
        cv2.waitKey(1)
    elif is_show:
        cv2.imshow('image', masked_image)
        cv2.waitKey(0)
    return masked_image


def _apply_mask2(image, mask, color, alpha=0.5):
    for c in range(3):
        image[:, :, c] = np.where(mask == 1, image[:, :, c] * (1 - alpha) + alpha * color[c],
                                  image[:, :, c])
    return image


# --------------- Drawing 3d box on image parts --------------
def draw_one_3d_box_cv2(img, box_3d, obj_id_name_map, score, tlwhy_format=False, calib_cam_to_img_p2=None,
                        force_color=None):
    """
    provide a obj id name map like: {1, 'car'}
    id to distinguish with previous object type

    tlwhy means input box are in format: [x, y, z, l, w, h, ry]
    that means we should convert it first.
    :param img:
    :param box_3d:
    :param obj_id_name_map:
    :param score:
    :param tlwhy_format:
    :param calib_cam_to_img_p2:
    :param force_color:
    :return:
    """
    assert isinstance(obj_id_name_map, dict), 'obj_id_name_map must be dict'
    # color = None
    if force_color:
        color = force_color
    else:
        color = create_unique_color_uchar(list(obj_id_name_map.keys())[0])
    if tlwhy_format:
        # transform [x, y, z, l, w, h, ry] to normal box
        assert calib_cam_to_img_p2, 'You should provide calibration matrix, convert camera to image coordinate.'
        center = box_3d[0: 3]
        dims = box_3d[3: 6]
        rot_y = -box_3d[6] / 180 * np.pi
        # alpha / 180 * np.pi + np.arctan(center[0] / center[2])

        converted_box_3d = []
        for i in [1, -1]:
            for j in [1, -1]:
                for k in [0, 1]:
                    point = np.copy(center)
                    point[0] = center[0] + i * dims[1] / 2 * np.cos(-rot_y + np.pi / 2) + \
                        (j * i) * dims[2] / 2 * np.cos(-rot_y)
                    point[2] = center[2] + i * dims[1] / 2 * np.sin(-rot_y + np.pi / 2) + \
                        (j * i) * dims[2] / 2 * np.sin(-rot_y)
                    point[1] = center[1] - k * dims[0]

                    point = np.append(point, 1)
                    point = np.dot(calib_cam_to_img_p2, point)
                    point = point[:2] / point[2]
                    point = point.astype(np.int16)
                    converted_box_3d.append(point)
        print('final box: ', converted_box_3d)
        # box_3d = np.asarray(converted_box_3d)
        box_3d = converted_box_3d
        # print(box_3d.shape)
        for i in range(4):
            point_1_ = box_3d[2 * i]
            point_2_ = box_3d[2 * i + 1]
            cv2.line(img, (point_1_[0], point_1_[1]),
                     (point_2_[0], point_2_[1]), color, 1)

        for i in range(8):
            point_1_ = box_3d[i]
            point_2_ = box_3d[(i + 2) % 8]
            cv2.line(img, (point_1_[0], point_1_[1]),
                     (point_2_[0], point_2_[1]), color, 1)
        return img
    else:
        # assert len(box_3d) == 8, 'every box 3d should have 8 points. if you got 7, you may want tlwhy=True'
        face_idx = np.array([0, 1, 5, 4,  # front face
                             1, 2, 6, 5,  # left face
                             2, 3, 7, 6,  # back face
                             3, 0, 4, 7]).reshape((4, 4))
        # print('start draw...')
        for i in range(4):
            x = np.append(box_3d[0, face_idx[i, ]],
                          box_3d[0, face_idx[i, 0]])
            y = np.append(box_3d[1, face_idx[i, ]],
                          box_3d[1, face_idx[i, 0]])
            # print('x: ', x)
            # print('y: ', y)
            # cv2.line(img, (point_1_, point_1_), (point_2_, point_2_), color, 1)
            pts = np.vstack((x, y)).T
            # filter negative values
            pts = (pts + abs(pts)) / 2
            pts = np.array([pts], dtype=int)
            # print(pts)
            cv2.polylines(img, pts, isClosed=True, color=color, thickness=1)
            if i == 3:
                # add text
                ori_txt = pts[0][1]

        return img



"""
HTML formatter.
"""

from markdown import markdown


class HTMLFormatter:
    """
    Convert lines in the HTML.
    """

    format = 'html'

    @staticmethod
    def write(lines):
        md = markdown(''.join(lines), output_format='html')
        return f'<html>\n<head></head>\n<body>\n{md}\n</body>\n</html>'.encode()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: labelmap.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='labelmap.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\x0elabelmap.proto\"M\n\x0cLabelMapItem\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05label\x18\x02 \x01(\x05\x12\n\n\x02id\x18\x04 \x01(\x05\x12\x14\n\x0c\x64isplay_name\x18\x03 \x01(\t\"\'\n\x08LabelMap\x12\x1b\n\x04item\x18\x01 \x03(\x0b\x32\r.LabelMapItem')
)




_LABELMAPITEM = _descriptor.Descriptor(
  name='LabelMapItem',
  full_name='LabelMapItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='LabelMapItem.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='label', full_name='LabelMapItem.label', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='id', full_name='LabelMapItem.id', index=2,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='LabelMapItem.display_name', index=3,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18,
  serialized_end=95,
)


_LABELMAP = _descriptor.Descriptor(
  name='LabelMap',
  full_name='LabelMap',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='item', full_name='LabelMap.item', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=97,
  serialized_end=136,
)

_LABELMAP.fields_by_name['item'].message_type = _LABELMAPITEM
DESCRIPTOR.message_types_by_name['LabelMapItem'] = _LABELMAPITEM
DESCRIPTOR.message_types_by_name['LabelMap'] = _LABELMAP
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LabelMapItem = _reflection.GeneratedProtocolMessageType('LabelMapItem', (_message.Message,), {
  'DESCRIPTOR' : _LABELMAPITEM,
  '__module__' : 'labelmap_pb2'
  # @@protoc_insertion_point(class_scope:LabelMapItem)
  })
_sym_db.RegisterMessage(LabelMapItem)

LabelMap = _reflection.GeneratedProtocolMessageType('LabelMap', (_message.Message,), {
  'DESCRIPTOR' : _LABELMAP,
  '__module__' : 'labelmap_pb2'
  # @@protoc_insertion_point(class_scope:LabelMap)
  })
_sym_db.RegisterMessage(LabelMap)


# @@protoc_insertion_point(module_scope)




def get_events_topic(cid):
    return "events/" + cid;

def get_presence_topic(cid):
    return "presence/" + cid

def get_personal_events_topic(cid):
    return "personalevents/" + cid

def get_chatting_topic(cid):
    return "messages/" + cid

def get_file_chatting_topic(cid):
    return "filemessages/" + cid

def get_archives_rooms_topic(cid):
    return "archivesrooms/" + cid

def get_archives_messages_topic(cid):
    return "archivesmessages/" + cid

def get_archives_myid_topic(cid):
    return "archivesmyid/" + cid



import numbers
import numpy as np
import six

from .dataset_mixin import DatasetMixin


def _is_iterable(x):
    if isinstance(x, str):
        return False
    return hasattr(x, '__iter__')


def _as_tuple(t):
    if _is_iterable(t):
        return tuple(t)
    else:
        return t,


def _bool_to_indices(indices, len_):
    true_indices = []
    for i, index in enumerate(indices):
        if isinstance(index, (bool, np.bool_)):
            if index:
                true_indices.append(i)
        else:
            return indices

    if not len(indices) == len_:
        raise ValueError(
            'The number of booleans is different from the length of dataset')
    return true_indices


def _as_key_indices(keys, key_names):
    key_names = _as_tuple(key_names)
    keys = _bool_to_indices(_as_tuple(keys), len(key_names))

    for key in keys:
        if isinstance(key, numbers.Integral):
            key_index = key
            if key_index < 0:
                key_index += len(key_names)
            if key_index not in range(0, len(key_names)):
                raise IndexError(
                    'index {} is out of bounds for keys with size {}'.format(
                        key, len(key_names)))
        else:
            try:
                key_index = key_names.index(key)
            except ValueError:
                raise KeyError('{} does not exists'.format(key))
        yield key_index


class SliceableDataset(DatasetMixin):
    """An abstract dataset class that supports slicing.

    This ia a dataset class that supports slicing.
    A dataset class inheriting this class should implement
    three methods: :meth:`__len__`, :meth:`keys`, and
    :meth:`get_example_by_keys`.

    Users can easily create sliceable datasets using
    :class:`~chainercv.chainer_experimental.datasets.sliceable.GetterDataset`
    or
    :class:`~chainercv.chainer_experimental.datasets.sliceable.TupleDataset`.
    """

    def __len__(self):
        raise NotImplementedError

    @property
    def keys(self):
        """Return names of all keys

        Returns:
            string or tuple of strings
        """
        raise NotImplementedError

    def get_example_by_keys(self, index, key_indices):
        """Return data of an example by keys

        Args:
            index (int): An index of an example.
            key_indices (tuple of ints): A tuple of indices of requested keys.

        Returns:
            tuple of data
        """
        raise NotImplementedError

    def get_example(self, index):
        if isinstance(self.keys, tuple):
            return self.get_example_by_keys(
                index, tuple(range(len(self.keys))))
        else:
            return self.get_example_by_keys(index, (0,))[0]

    @property
    def slice(self):
        return SliceHelper(self)

    def __iter__(self):
        return (self.get_example(i) for i in six.moves.range(len(self)))


class SliceHelper(object):
    """A helper class for :class:`SliceableDataset`."""

    def __init__(self, dataset):
        self._dataset = dataset

    def __getitem__(self, args):
        if isinstance(args, tuple):
            indices, keys = args
        else:
            indices = args
            keys = self._dataset.keys

        if not isinstance(indices, slice):
            indices = _bool_to_indices(indices, len(self._dataset))
        key_indices = tuple(_as_key_indices(keys, self._dataset.keys))
        return_tuple = _is_iterable(keys)

        return SlicedDataset(
            self._dataset, indices,
            tuple(key_indices) if return_tuple else key_indices[0])


class SlicedDataset(SliceableDataset):
    """A sliced view for :class:`SliceableDataset`."""

    def __init__(self, dataset, indices, key_indices):
        self._dataset = dataset
        self._indices = indices
        self._key_indices = key_indices

    def __len__(self):
        if isinstance(self._indices, slice):
            start, end, step = self._indices.indices(len(self._dataset))
            return len(range(start, end, step))
        else:
            return len(self._indices)

    @property
    def keys(self):
        keys = _as_tuple(self._dataset.keys)
        if isinstance(self._key_indices, tuple):
            return tuple(keys[key_index] for key_index in self._key_indices)
        else:
            return keys[self._key_indices]

    def get_example_by_keys(self, index, key_indices):
        if isinstance(key_indices, tuple):
            key_indices = tuple(
                _as_tuple(self._key_indices)[key_index]
                for key_index in key_indices)
        else:
            key_indices = _as_tuple(self._key_indices)[key_indices]

        if isinstance(self._indices, slice):
            start, _, step = self._indices.indices(len(self._dataset))
            return self._dataset.get_example_by_keys(
                start + index * step, key_indices)
        else:
            return self._dataset.get_example_by_keys(
                self._indices[index], key_indices)



#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""

this script will using pycoco API
draw our converted annotation to check
if result is right or not

"""
try:
    from pycocotools.coco import COCO
    from pycocotools import mask as maskUtils
except ImportError as e:
    print('Got import error: {}'.format(e))
    print('[WARN] you are not either install pycocotools or its dependencies. pls install first.')
    # exit(-1)
import os
import sys
import cv2
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection
from matplotlib.patches import Polygon
from alfred.utils.log import logger as logging
import cv2
from alfred.vis.image.det import visualize_det_cv2_part
from alfred.vis.image.common import get_unique_color_by_id
import numpy as np
from pprint import pprint


# USED_CATEGORIES_IDS = [i for i in range(1, 16)]

def get_random_color():
    return list(np.array(np.random.random(size=3) * 256).astype(np.uint8))


def showAnns(ori_img, anns, draw_bbox=False):
    h, w, c = ori_img.shape
    if len(anns) == 0:
        return ori_img
    if 'segmentation' in anns[0] or 'keypoints' in anns[0]:
        datasetType = 'instances'
    elif 'caption' in anns[0]:
        datasetType = 'captions'
    else:
        raise Exception('datasetType not supported')
    if datasetType == 'instances':
        mask = np.zeros_like(ori_img).astype(np.uint8)

        for ann in anns:
            c = np.array((np.random.random((1, 3)) * 0.6 +
                          0.4)[0]*255).astype(int).tolist()
            if 'segmentation' in ann:
                if type(ann['segmentation']) == list:
                    # polygon
                    for seg in ann['segmentation']:
                        poly = np.array(seg).reshape(
                            (int(len(seg) / 2), 2))
                        pts = poly.reshape((-1, 1, 2))
                        cv2.polylines(
                            ori_img, np.int32([pts]), True, c, thickness=1, lineType=cv2.LINE_AA)
                        cv2.drawContours(mask, np.int32([pts]), -1, c, -1)

                        if cv2.contourArea(np.int32(pts)) > 1:
                            M = cv2.moments(np.int32(pts))
                            cX = int(M["m10"] / M["m00"])
                            cY = int(M["m01"] / M["m00"])
                            cv2.putText(ori_img, 'CAT:{}'.format(
                                ann['category_id']), (cX, cY), cv2.FONT_HERSHEY_PLAIN, 0.8, (255, 255, 255), 1, cv2.LINE_AA)
                else:
                    # mask
                    if type(ann['segmentation']['counts']) == list:
                        rle = maskUtils.frPyObjects([ann['segmentation']],
                                                    h, w)
                    else:
                        rle = [ann['segmentation']]
                    m = maskUtils.decode(rle)
                    img = np.ones((m.shape[0], m.shape[1], 3)).astype(np.uint8)
                    if ann['iscrowd'] == 1:
                        color_mask = np.array([2.0, 166.0, 101.0])
                    if ann['iscrowd'] == 0:
                        color_mask = np.random.random((1, 3)).tolist()[0]
                    img *= get_random_color()
                    img = cv2.bitwise_or(img, img, mask=m)
                    mask += img
            if draw_bbox:
                if 'bbox' in ann.keys():
                    [bbox_x, bbox_y, bbox_w, bbox_h] = ann['bbox']
                    pt1 = (int(bbox_x), int(bbox_y))
                    pt2 = (int(bbox_x+bbox_w), int(bbox_y+bbox_h))
                    cv2.rectangle(ori_img, pt1, pt2, color=c,
                                  thickness=1, lineType=cv2.LINE_AA)

            if 'keypoints' in ann and type(ann['keypoints']) == list:
                # turn skeleton into zero-based index
                # sks = np.array(
                #     self.loadCats(ann['category_id'])[0]['skeleton']) - 1
                kp = np.array(ann['keypoints'])
                x = kp[0::3]
                y = kp[1::3]
                v = kp[2::3]
                # for sk in sks:
                #     if np.all(v[sk] > 0):
                #         cv2.line(ori_img, x[sk], y[sk], color=c)
                print(kp)
                print('keypoint vis not supported')

        if type(ann['segmentation']) == list:
            ori_img = cv2.addWeighted(ori_img, 0.7, mask, 0.6, 0.7)
        else:
            print('[WARN] you are using RLE mask encode format.')
            ori_img = cv2.addWeighted(ori_img, 0.7, mask, 0.6, 0.7)
    elif datasetType == 'captions':
        for ann in anns:
            print(ann['caption'])
    return ori_img


def vis_coco(coco_img_root, ann_f):
    data_dir = coco_img_root
    coco = COCO(ann_f)

    cats = coco.loadCats(coco.getCatIds())
    logging.info('cats: {}'.format(cats))
    cats_new = dict()
    for c in cats:
        cats_new[c['id']] = c['name']
    pprint(cats_new)
    pprint([i['name'] for i in cats])
    print('All {} classes.'.format(len(cats_new)))
    img_ids = coco.getImgIds()
    logging.info('all images we got: {}'.format(len(img_ids)))

    # draw instances
    for img_id in img_ids:
        img = coco.loadImgs(img_id)[0]
        print('checking img: {}, id: {}'.format(img, img_id))

        # img['file_name'] may be not basename
        if 'file_name' in img.keys():
            img_f = os.path.join(data_dir, os.path.basename(img['file_name']))
        elif 'filename' in img.keys():
            img_f = os.path.join(data_dir, os.path.basename(img['filename']))
        else:
            print(
                'does not foud a file_name or filename in feild. check your annotation style: ', img)
        assert(os.path.exists(
            img_f)), '{} not found, maybe your filename pattern not right? Pls fire an issue to alfred github repo!'.format(img_f)
        anno_ids = coco.getAnnIds(imgIds=img['id'])
        annos = coco.loadAnns(anno_ids)

        logging.info('showing anno: {} objects. '.format(len(annos)))
        if len(annos) > 0 and len(annos[0]['segmentation']) == 0:
            logging.info('no segmentation found, using opencv vis.')
            img = cv2.imread(img_f)

            font = cv2.FONT_HERSHEY_SIMPLEX
            font_scale = 0.36
            font_thickness = 1
            line_thickness = 1

            for ann in annos:
                b = ann['bbox']
                x1 = int(b[0])
                y1 = int(b[1])
                x2 = int(x1 + b[2])
                y2 = int(y1 + b[3])
                cls_id = ann['category_id']
                unique_color = get_unique_color_by_id(cls_id)
                cv2.rectangle(img, (x1, y1), (x2, y2),
                              unique_color, line_thickness, cv2.LINE_AA)
                if cls_id > len(cats_new):
                    print('WARN: seems your category id not same with your meta info!!')
                text_label = '{}:{}'.format(cls_id, cats_new[cls_id])
                (ret_val, _) = cv2.getTextSize(
                    text_label, font, font_scale, font_thickness)
                txt_bottom_left = (x1+4, y1-4)
                cv2.rectangle(img, (txt_bottom_left[0]-4, txt_bottom_left[1] - ret_val[1]-2),
                              (txt_bottom_left[0] + ret_val[0] +
                               2, txt_bottom_left[1]+4),
                              (0, 0, 0), -1)
                cv2.putText(img, text_label, txt_bottom_left, font,
                            font_scale, (237, 237, 237), font_thickness, cv2.LINE_AA)
            cv2.imshow('rr', img)
            cv2.waitKey(0)
        else:
            im = cv2.imread(img_f)
            # plt.imshow(I)
            # plt.axis('off')
            # coco.showAnns(annos, True)
            # plt.show()
            ori_im = showAnns(im, annos, True)
            if ori_im is not None:
                cv2.imshow('aa', ori_im)
                cv2.waitKey(0)
            else:
                I = Image.open(img_f)
                plt.imshow(I)
                plt.axis('off')
                coco.showAnns(annos, True)
                plt.show()



from collections import defaultdict
import json
import numpy as np
import os
import PIL.Image
import PIL.ImageDraw

from ..meta.getter_dataset import GetterDataset
import cv2

try:
    from pycocotools import mask as coco_mask
except ImportError:
    pass


class COCOInstancesBaseDataset(GetterDataset):

    def __init__(self, data_dir='auto', split='train', year='2017',
                 use_crowded=False, read_img_with='pil'):
        if year == '2017' and split in ['minival', 'valminusminival']:
            raise ValueError(
                'coco2017 dataset does not support given split: {}'
                .format(split))

        super(COCOInstancesBaseDataset, self).__init__()
        self.use_crowded = use_crowded
        self.read_img_with = read_img_with

        if split in ['val', 'minival', 'valminusminival']:
            img_split = 'val'
        else:
            img_split = 'train'

        self.img_root = os.path.join(
            data_dir, '{}{}'.format(img_split, year))
        tmp = self.img_root
        if not os.path.exists(self.img_root):
            self.img_root = os.path.join(data_dir, 'images')
        assert os.path.exists(
            self.img_root), 'we have tried img_root: {} and {}, either found, did u make sure it exists?'.format(tmp, self.img_root)
        anno_path = os.path.join(
            data_dir, 'annotations', 'instances_{}{}.json'.format(split, year))

        self.data_dir = data_dir
        annos = json.load(open(anno_path, 'r'))

        self.id_to_prop = {}
        for prop in annos['images']:
            self.id_to_prop[prop['id']] = prop
        self.ids = sorted(list(self.id_to_prop.keys()))

        self.cat_ids = [cat['id'] for cat in annos['categories']]

        self.id_to_anno = defaultdict(list)
        for anno in annos['annotations']:
            self.id_to_anno[anno['image_id']].append(anno)

        self.add_getter('img', self._get_image)
        self.add_getter('mask', self._get_mask)
        self.add_getter(
            ['bbox', 'label', 'area', 'crowded'],
            self._get_annotations)

    def __len__(self):
        return len(self.ids)

    def _get_image(self, i):
        img_path = os.path.join(
            self.img_root, self.id_to_prop[self.ids[i]]['file_name'])
        if self.read_img_with == 'cv2':
            img = cv2.imread(img_path)
        else:
            img = PIL.Image.open(img_path)
        return img

    def _get_mask(self, i):
        # List[{'segmentation', 'area', 'iscrowd',
        #       'image_id', 'bbox', 'category_id', 'id'}]
        annotation = self.id_to_anno[self.ids[i]]
        H = self.id_to_prop[self.ids[i]]['height']
        W = self.id_to_prop[self.ids[i]]['width']

        mask = []
        crowded = []
        for anno in annotation:
            msk = self._segm_to_mask(anno['segmentation'], (H, W))
            # FIXME: some of minival annotations are malformed.
            if msk.shape != (H, W):
                continue
            mask.append(msk)
            crowded.append(anno['iscrowd'])
        mask = np.array(mask, dtype=np.bool)
        crowded = np.array(crowded, dtype=np.bool)
        if len(mask) == 0:
            mask = np.zeros((0, H, W), dtype=np.bool)

        if not self.use_crowded:
            not_crowded = np.logical_not(crowded)
            mask = mask[not_crowded]
        return mask

    def _get_annotations(self, i):
        # List[{'segmentation', 'area', 'iscrowd',
        #       'image_id', 'bbox', 'category_id', 'id'}]
        annotation = self.id_to_anno[self.ids[i]]
        bbox = np.array([ann['bbox'] for ann in annotation],
                        dtype=np.float32)
        if len(bbox) == 0:
            bbox = np.zeros((0, 4), dtype=np.float32)
        # (x, y, width, height)  -> (x_min, y_min, x_max, y_max)
        bbox[:, 2] = bbox[:, 0] + bbox[:, 2]
        bbox[:, 3] = bbox[:, 1] + bbox[:, 3]
        # (x_min, y_min, x_max, y_max) -> (y_min, x_min, y_max, x_max)
        bbox = bbox[:, [1, 0, 3, 2]]

        label = np.array([self.cat_ids.index(ann['category_id'])
                          for ann in annotation], dtype=np.int32)

        area = np.array([ann['area']
                         for ann in annotation], dtype=np.float32)

        crowded = np.array([ann['iscrowd']
                            for ann in annotation], dtype=np.bool)

        # Remove invalid boxes
        bbox_area = np.prod(bbox[:, 2:] - bbox[:, :2], axis=1)
        keep_mask = np.logical_and(bbox[:, 0] <= bbox[:, 2],
                                   bbox[:, 1] <= bbox[:, 3])
        keep_mask = np.logical_and(keep_mask, bbox_area > 0)

        if not self.use_crowded:
            keep_mask = np.logical_and(keep_mask, np.logical_not(crowded))

        bbox = bbox[keep_mask]
        label = label[keep_mask]
        area = area[keep_mask]
        crowded = crowded[keep_mask]
        return bbox, label, area, crowded

    def _segm_to_mask(self, segm, size):
        # Copied from pycocotools.coco.COCO.annToMask
        H, W = size
        if isinstance(segm, list):
            # polygon -- a single object might consist of multiple parts
            # we merge all parts into one mask rle code
            mask = np.zeros((H, W), dtype=np.uint8)
            mask = PIL.Image.fromarray(mask)
            for sgm in segm:
                xy = np.array(sgm).reshape((-1, 2))
                xy = [tuple(xy_i) for xy_i in xy]
                PIL.ImageDraw.Draw(mask).polygon(xy=xy, outline=1, fill=1)
            mask = np.asarray(mask)
        elif isinstance(segm['counts'], list):
            rle = coco_mask.frPyObjects(segm, H, W)
            mask = coco_mask.decode(rle)
        else:
            mask = coco_mask.decode(segm)
        return mask.astype(np.bool)


class COCOBboxDataset(COCOInstancesBaseDataset):

    def __init__(self, data_dir='auto', split='train', year='2017',
                 use_crowded=False, read_img_with='pil', return_area=False, return_crowded=False):
        super(COCOBboxDataset, self).__init__(
            data_dir, split, year, use_crowded, read_img_with)
        keys = ('img', 'bbox', 'label')
        if return_area:
            keys += ('area',)
        if return_crowded:
            keys += ('crowded',)
        self.keys = keys



#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""

view txt labeled detection data


"""
import os
import sys
import cv2
from glob import glob
import os
import sys
import cv2
from alfred.utils.log import logger as logging


def vis_det_yolo(img_root, label_root):
    logging.info('img root: {}, label root: {}'.format(img_root, label_root))
    # auto detection .jpg or .png images
    txt_files = glob(os.path.join(label_root, '*.txt'))
    for txt_f in txt_files:
        img_f = os.path.join(img_root, os.path.basename(txt_f).split('.')[0] + '.jpg')
        if os.path.exists(img_f):
            img = cv2.imread(img_f)
            h, w, _ = img.shape
            if os.path.exists(txt_f):
                with open(txt_f) as f:
                    annos = f.readlines() 
                    for ann in annos:
                        ann = ann.strip().split(' ')
                        category = ann[0]
                        x = float(ann[1]) * w
                        y = float(ann[2]) * h
                        bw = float(ann[3]) * w
                        bh = float(ann[4]) * h
                        xmin = int(x - bw/2)
                        ymin = int(y - bh/2)
                        xmax = int(x + bw/2)
                        ymax = int(y + bh/2)
                        print(xmin, ymin, xmax, ymax, category)
                        cv2.putText(img, category, (xmin, ymin), cv2.FONT_HERSHEY_COMPLEX, 0.7, (255, 255, 255))
                        cv2.rectangle(img, (xmin, ymin), (xmax, ymax), (0, 255, 0), 2, 1)
                cv2.imshow('yolo check', img)
                cv2.waitKey(0)
            else:
                logging.warning('xxxx image: {} not found.'.format(img_f))


if __name__ == "__main__":
    vis_det_txt(sys.argv[1], sys.argv[2])



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from utils.mana import welcome

from utils.log import logger as logging
from vis.image.det import visualize_det_cv2
import cv2
import numpy as np
from vis.image.get_dataset_label_map import coco_label_map_list
from vis.image.common import draw_rect_with_style
import torch
from dl.torch.common import print_tensor

from varname import varname


def a_func(num):
    print(varname() + ': ' + str(num))


def clothes(func):
    def wear():
        print('Buy clothes!{}'.format(func.__name__))
        return func()
    return wear


@clothes
def body():
    print('The body feels could!')


if __name__ == '__main__':
    v = a_func(1098)

    # welcome('')
    # logging.info('hi hiu')
    # logging.error('ops')

    # a = cv2.imread('/home/jintian/Pictures/1.jpeg')

    # dets = [
    #     [1, 0.9, 4, 124, 333, 256],
    #     [2, 0.7, 155, 336, 367, 485],
    # ]
    # dets = np.array(dets)
    # print(type(a))

    # draw_rect_with_style(a, (78, 478), (478, 223), (0, 255, 255), style='dashed')
    # visualize_det_cv2(a, dets, coco_label_map_list, is_show=True)

    aaa = torch.randn([1, 23, 45])
    print_tensor(aaa)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import sys
try:
    import cv2
except ImportError:
    print('you are not install opencv-python, using pip install opencv-python install it.')
from colorama import Fore, Back, Style


class VideoExtractor(object):

    def __init__(self, jump_frames=6, save_format='frame_%06d.jpg'):
        """
        we set frames to jump, etc, using jump_frames=6
        will save one frame per 6 frames jumped
        :param jump_frames:
        :param save_format: this is the frames save format
        users can decide what's the format is: frame_0000004.jpg
        """
        self.current_frame = 0
        self.current_save_frame = 0
        if jump_frames:
            self.jump_frames = int(jump_frames)
        else:
            self.jump_frames = 6
        self.save_format = save_format

    def extract(self, video_f):
        if os.path.exists(video_f) and os.path.isfile(video_f):
            cap = cv2.VideoCapture(video_f)
            f_n = os.path.basename(video_f).split('.')[0]

            save_dir = os.path.join(os.path.dirname(video_f), os.path.basename(video_f).split('.')[0])
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)

            res = True
            while res:
                res, image = cap.read()
                self.current_frame += 1
                if self.current_frame % self.jump_frames == 0:
                    print('Read frame: {} jump frames: {}'.format(self.current_frame, self.jump_frames))
                    cv2.imwrite(os.path.join(save_dir, f_n + '_' + self.save_format % self.current_save_frame), image)
                    self.current_save_frame += 1

            print(Fore.GREEN + Style.BRIGHT)
            print('Success!')
        else:
            print(Fore.RED + Style.BRIGHT)
            print('Error! ' + Style.RESET_ALL + '{} not exist.'.format(video_f))



# Copyright (c) 2021, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
from PIL import Image
import cv2
import torch
from torch import nn
import torchvision.transforms as T

import pycuda.driver as cuda
import pycuda.autoinit

try:
    import cupy as cp
except ImportError:
    print('You should install cupy for preprocess with CUDA: https://docs.cupy.dev/en/stable/install.html')
# from cupy.core.dlpack import toDlpack
# from cupy.core.dlpack import fromDlpack
from torch.utils.dlpack import to_dlpack
from torch.utils.dlpack import from_dlpack
from alfred.dl.torch.common import device


def preprocess_np(img_path):
    '''process use numpy
    '''
    im = Image.open(img_path)
    img = im.resize((800, 800), Image.BILINEAR)
    img = np.array(img).astype(np.float32) / 255.0
    img = img.transpose(2, 0, 1)
    # print(img.shape)
    img = (img - np.array([[[0.485]], [[0.456]], [[0.406]]])
           )/np.array([[[0.229]], [[0.224]], [[0.225]]])

    # img = img.transpose(1,2,0)
    img = np.expand_dims(img, axis=0)
    img = np.ascontiguousarray(img)
    img = np.array(img).astype(np.float32)

    return img, im, im.size


class PyTorchTensorHolder(pycuda.driver.PointerHolderBase):
    '''代码来源:
        https://github.com/NVIDIA/trt-samples-for-hackathon-cn/blob/master/python/app_onnx_resnet50.py
    '''

    def __init__(self, tensor):
        super(PyTorchTensorHolder, self).__init__()
        self.tensor = tensor

    def get_pointer(self):
        return self.tensor.data_ptr()


transform = T.Compose([
    T.Resize((800, 800)),  # PIL.Image.BILINEAR
    T.ToTensor(),
    # T.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
])


def preprocess_torch(img_path):
    '''process use torchvision
    '''
    im = Image.open(img_path)
    img = transform(im).unsqueeze(0)
    img = PyTorchTensorHolder(img)
    return img, im, im.size


def preprocess_torch_v1(img_path):
    im = Image.open(img_path)
    img = transform(im).unsqueeze(0).cpu().numpy()
    return img, im, im.size


def preprocess_np_no_normalize(img_path):
    im = cv2.imread(img_path)
    print(img_path)
    print(im.shape)
    # img = transform(im).unsqueeze(0)
    a = np.transpose(im, (2, 0, 1)).astype(np.float32)
    return a, im


def preprocess_cu(img_np):
    mean_cp = cp.array([[[0.485]], [[0.456]], [[0.406]]])
    std_cp = cp.array([[[0.229]], [[0.224]], [[0.225]]])

    img_cu = cp.divide(cp.asarray(img, dtype=cp.float32), 255.0)
    img_cu = img_cu.transpose(2, 0, 1)
    img_cu = cp.subtract(img_cu, mean_cp)
    img_cu = cp.divide(img_cu, std_cp)

    # cupy to torch tensor
    # img_tensor = from_dlpack(toDlpack(img_cu))

    return img_cu



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import importlib
import importlib.util
import logging
import os
import random
import sys
from datetime import datetime

import numpy as np
import torch

__all__ = ["seed_all_rng"]


def seed_all_rng(seed=None):
    """
    Set the random seed for the RNG in torch, numpy and python.

    Args:
        seed (int): if None, will use a strong random seed.
    """
    if seed is None:
        seed = (
            os.getpid()
            + int(datetime.now().strftime("%S%f"))
            + int.from_bytes(os.urandom(2), "big")
        )
        logger = logging.getLogger(__name__)
        logger.info("Using a generated random seed {}".format(seed))
    np.random.seed(seed)
    torch.set_rng_state(torch.manual_seed(seed).get_state())
    random.seed(seed)


# from https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path
def _import_file(module_name, file_path, make_importable=False):
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    if make_importable:
        sys.modules[module_name] = module
    return module


def _configure_libraries():
    """
    Configurations for some libraries.
    """
    # An environment option to disable `import cv2` globally,
    # in case it leads to negative performance impact
    disable_cv2 = int(os.environ.get("dl_lib_DISABLE_CV2", False))
    if disable_cv2:
        sys.modules["cv2"] = None
    else:
        # Disable opencl in opencv since its interaction with cuda often has negative effects
        # This envvar is supported after OpenCV 3.4.0
        os.environ["OPENCV_OPENCL_RUNTIME"] = "disabled"
        try:
            import cv2

            if int(cv2.__version__.split(".")[0]) >= 3:
                cv2.ocl.setUseOpenCL(False)
        except ImportError:
            pass


_ENV_SETUP_DONE = False


def setup_environment():
    """Perform environment setup work. The default setup is a no-op, but this
    function allows the user to specify a Python source file or a module in
    the $dl_lib_ENV_MODULE environment variable, that performs
    custom setup work that may be necessary to their computing environment.
    """
    global _ENV_SETUP_DONE
    if _ENV_SETUP_DONE:
        return
    _ENV_SETUP_DONE = True

    _configure_libraries()

    custom_module_path = os.environ.get("dl_lib_ENV_MODULE")

    if custom_module_path:
        setup_custom_environment(custom_module_path)
    else:
        # The default setup is a no-op
        pass


def setup_custom_environment(custom_module):
    """
    Load custom environment setup by importing a Python source file or a
    module, and run the setup function.
    """
    if custom_module.endswith(".py"):
        module = _import_file("dl_lib.utils.env.custom_module", custom_module)
    else:
        module = importlib.import_module(custom_module)
    assert hasattr(module, "setup_environment") and callable(module.setup_environment), (
        "Custom environment module defined in {} does not have the "
        "required callable attribute 'setup_environment'."
    ).format(custom_module)
    module.setup_environment()





"""

Split coco dataset

"""
import json
import argparse
import funcy
from sklearn.model_selection import train_test_split
import os


def save_coco(file, info, licenses, images, annotations, categories):
    with open(file, 'wt', encoding='UTF-8') as coco:
        json.dump({'info': info, 'licenses': licenses, 'images': images,
                   'annotations': annotations, 'categories': categories}, coco, indent=2, sort_keys=True)


def filter_annotations(annotations, images):
    image_ids = funcy.lmap(lambda i: int(i['id']), images)
    return funcy.lfilter(lambda a: int(a['image_id']) in image_ids, annotations)


def split_coco(ann_f, split=0.8, remove_empty=False):
    train_f = os.path.join(os.path.dirname(ann_f), os.path.basename(
        ann_f).replace('.json', '_train.json'))
    val_f = os.path.join(os.path.dirname(ann_f), os.path.basename(
        ann_f).replace('.json', '_val.json'))
    with open(ann_f, 'rt', encoding='UTF-8') as annotations:
        coco = json.load(annotations)
        info = ''
        if 'info' in coco.keys():
            info = coco['info']
        licenses = ''
        if 'licenses' in coco.keys():
            licenses = coco['licenses']
        images = coco['images']
        annotations = coco['annotations']
        categories = coco['categories']

        number_of_images = len(images)

        if remove_empty:
            images_with_annotations = funcy.lmap(
                lambda a: int(a['image_id']), annotations)

            # filter out images without annotations
            images = funcy.lremove(
                lambda i: i['id'] not in images_with_annotations, images)
            print('removed {} images without annotations, all images: {}, now: {}'.format(
                number_of_images-len(images), number_of_images, len(images)
            ))
        else:
            print('all images: {}'.format(number_of_images))

        x, y = train_test_split(images, train_size=float(split))

        save_coco(train_f, info, licenses, x,
                  filter_annotations(annotations, x), categories)
        save_coco(val_f, info, licenses, y,
                  filter_annotations(annotations, y), categories)

        print("Saved {} entries in {} and {} in {}.".format(
            len(x), train_f, len(y), val_f))



"""

Process on image

"""


import cv2


def darken_image(ori_img, dark_factor=0.6):
    """
    this will darken origin image and return darken one
    """
    hsv_img = cv2.cvtColor(ori_img, cv2.COLOR_BGR2HSV)
    hsv_img[...,2] = hsv_img[...,2]*dark_factor
    originimg = cv2.cvtColor(hsv_img, cv2.COLOR_HSV2BGR)
    return originimg


#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import ctypes
import math
import time
import torch


def scatter_nd(indices, updates, shape):
    """pytorch edition of tensorflow scatter_nd.
    this function don't contain except handle code. so use this carefully
    when indice repeats, don't support repeat add which is supported
    in tensorflow.
    """
    ret = torch.zeros(*shape, dtype=updates.dtype, device=updates.device)
    ndim = indices.shape[-1]
    output_shape = list(indices.shape[:-1]) + shape[indices.shape[-1]:]
    flatted_indices = indices.view(-1, ndim)
    slices = [flatted_indices[:, i] for i in range(ndim)]
    slices += [Ellipsis]
    ret[slices] = updates.view(*output_shape)
    return ret


def gather_nd(params, indices):
    # this function has a limit that MAX_ADVINDEX_CALC_DIMS=5
    ndim = indices.shape[-1]
    output_shape = list(indices.shape[:-1]) + list(params.shape[indices.shape[-1]:])
    flatted_indices = indices.view(-1, ndim)
    slices = [flatted_indices[:, i] for i in range(ndim)]
    slices += [Ellipsis]
    return params[slices].view(*output_shape)



"""

stack images in matrix style

"""
import cv2
import numpy as np
from alfred.utils.log import logger as logging



def check_shape_resize_if_possible(imgs):
    shapes = [i.shape for i in imgs]
    if len(set(shapes))==1:
        return imgs
    else:
        logging.info('detected images shape not equal, resize to the first shape...')
        imgs = [cv2.resize(i, (shapes[0][1], shapes[0][0])) for i in imgs]
        return imgs



def stack_imgs(imgs_list, dim2d):
    """
    send a list of images
    then using dim2d to stack it

    for example:
        a.png
        b.png
        c.png
        d.png
    
    dim2d:
        2x2
    """
    a = int(dim2d.split('x')[0])
    b = int(dim2d.split('x')[1])
    if len(imgs_list) % a != 0 or len(imgs_list) % b:
        logging.info('dim2d {} is not applicable for {} images.'.format(dim2d, len(imgs_list)))
        exit(0)
    elif len(imgs_list) != a*b:
        logging.error('len imgs not equal to: axb={}'.format(a*b))
        exit(0)
    else:
        imgs_list = [cv2.imread(i) for i in imgs_list]
        all_raws = []
        # 2x1 bug?
        for ri in range(a):
            one_raw = []
            for ci in range(b):
                one_raw.append(imgs_list[ri*b + ci])
                logging.info('stacking row: {}, with len: {}'.format(ri, len(one_raw)))
            imgs = check_shape_resize_if_possible(one_raw)
            img_a = np.hstack(imgs)
            all_raws.append(img_a)
        all_raws = check_shape_resize_if_possible(all_raws)
        final_img = np.vstack(all_raws)
        logging.info('final combined img shape: {}'.format(final_img.shape))
        cv2.imwrite('stacked_img.jpg', final_img)
        logging.info('done.')




"""
Images extractor from markdown document.
"""

import markdown
from markdown.treeprocessors import Treeprocessor
from markdown.extensions import Extension
from markdown.inlinepatterns import SimpleTagPattern
from typing import List


__all__ = ['ArticleTransformer']


class ImgExtractor(Treeprocessor):
    def run(self, doc):
        """
        Find all images and append to markdown.images.
        """

        self.md.images = []
        for image in doc.findall('.//img'):
            self.md.images.append(image.get('src'))


class ImgExtExtension(Extension):
    def extendMarkdown(self, md, md_globals):
        img_ext = ImgExtractor(md)
        md.treeprocessors.register(img_ext, 'imgext', 20)


class ArticleTransformer:
    """
    Markdown article transformation class.
    """

    def __init__(self, article_path: str, image_downloader):
        self._image_downloader = image_downloader
        self._article_file_path = article_path
        self._md_conv = markdown.Markdown(extensions=[ImgExtExtension(), 'md_in_html'])
        self._replacement_mapping = {}

    def _read_article(self) -> List[str]:
        with open(self._article_file_path, 'r', encoding='utf8') as m_file:
            self._md_conv.convert(m_file.read())

        print(f'Images links count = {len(self._md_conv.images)}')
        images = set(self._md_conv.images)
        print(f'Unique images links count = {len(images)}')

        return images

    def _fix_document_urls(self) -> List[str]:
        # print('Replacing images urls in the document...')
        replacement_mapping = self._replacement_mapping
        lines = []
        with open(self._article_file_path, 'r', encoding='utf8') as infile:
            for line in infile:
                for src, target in replacement_mapping.items():
                    line = line.replace(src, target)
                lines.append(line)

        return lines

    def run(self):
        """
        Run article conversion.
        """

        self._replacement_mapping = self._image_downloader.download_images(self._read_article())
        return self._fix_document_urls()



import os
import json
import glob
from alfred.utils.log import logger
from PIL import Image
import numpy as np
import cv2
import shutil
from tqdm import tqdm


__all__ = ['YoloEvaluator']


"""

Parsing any dataset in Yolo format.
You can simply change classes names to yours.
"""

img_formats = ['bmp', 'jpg', 'jpeg', 'png', 'tif',
               'tiff', 'dng']  # acceptable image suffixes


class YoloEvaluator:

    def __init__(self,
                 imgs_root,
                 labels_root,
                 infer_func,
                 model=None,
                 prep_func=None,
                 posp_func=None,
                 conf_thr=0.4,
                 iou_thr=0.5):
        assert os.path.exists(imgs_root)
        assert os.path.exists(labels_root)

        self.infer_func = infer_func
        self.model_ = model
        self.prep_func = prep_func
        self.posp_func = posp_func

        self.conf_thr = conf_thr
        self.iou_thr = iou_thr

        self._data_root = imgs_root
        logger.info('data_root: {}'.format(self._data_root))

        self.img_files = []
        self.label_files = []
        self.load_combined_imgs_and_labels(imgs_root, labels_root)

        self.hold_vis = True
        logger.info(
            'Press space to vis image, press q to skip and continue eval.')

    def load_combined_imgs_and_labels(self, imgs_root, labels_root):
        logger.info('labels root path: {}, img root: {}, pls check it they right or not.'.format(
            imgs_root, labels_root))
        self.label_files = glob.glob(os.path.join(labels_root, "*.txt"))
        for ext in img_formats:
            self.img_files.extend(
                glob.glob(os.path.join(imgs_root, '*.{}'.format(ext))))
        if len(self.img_files) != len(self.label_files):
            imgs_names = [os.path.basename(i).split('.')[0]
                          for i in self.img_files]
            labels_names = [os.path.basename(i).split(
                '.')[0] for i in self.label_files]
            valid_files = [i for i in imgs_names if i in labels_names]
            logger.info('original imgs: {}, original labels: {}, valid num: {}'.format(
                len(self.img_files), len(self.label_files), len(valid_files)))
            # labels is more than images
            self.label_files = [os.path.join(
                labels_root, i + '.txt') for i in valid_files]
            self.img_files = [os.path.join(
                imgs_root, i + '.jpg') for i in valid_files]
        self.img_files = sorted(self.img_files)
        self.label_files = sorted(self.label_files)
        logger.info('img num: {}, label num: {}'.format(
            len(self.img_files), len(self.label_files)))
        logger.info(self.img_files[89])
        logger.info(self.label_files[89])
        logger.info('Please check if images and labels are paired properly.')

    def _load_yolo_boxes_and_labels(self, lb_file, ori_w, ori_h):
        with open(lb_file, 'r') as f:
            l = np.array([x.split() for x in f.read().strip(
            ).splitlines()], dtype=np.float32)  # labels
            l = np.clip(l, 0, 1.)
        res_boxes = []
        res_labels = []
        if len(l):
            boxes = l[:, 1:]
            labels = l[:, 0]

            for i, b in enumerate(boxes):
                cx, cy, w, h = b
                cx *= ori_w
                cy *= ori_h
                w *= ori_w
                h *= ori_h
                x = cx - w/2
                y = cy - h/2
                if x < 0 or y < 0 or w <= 2 or h <= 2:
                    continue
                res_boxes.append(np.array([x, y, x+w, y+h]).astype(np.int))
                res_labels.append(labels[i])
        return res_boxes, res_labels

    def exif_size(self, img):
        # Returns exif-corrected PIL size
        s = img.size  # (width, height)
        try:
            rotation = dict(img._getexif().items())[orientation]
            if rotation == 6:  # rotation 270
                s = (s[1], s[0])
            elif rotation == 8:  # rotation 90
                s = (s[1], s[0])
        except:
            pass
        return s

    @staticmethod
    def compute_iou(rec1, rec2):
        """
        computing IoU
        :param rec1: (x0, y0, x1, y1), which reflects
                (top, left, bottom, right)
        :param rec2: (x0, y0, x1, y1)
        :return: scala value of IoU
        """
        # computing area of each rectangles
        S_rec1 = (rec1[3] - rec1[1]) * (rec1[2] - rec1[0])
        S_rec2 = (rec2[3] - rec2[1]) * (rec2[2] - rec2[0])

        # computing the sum_area
        sum_area = S_rec1 + S_rec2

        # find the each edge of intersect rectangle
        left_line = max(rec1[0], rec2[0])
        right_line = min(rec1[2], rec2[2])
        top_line = max(rec1[1], rec2[1])
        bottom_line = min(rec1[3], rec2[3])

        # judge if there is an intersect
        if left_line >= right_line or top_line >= bottom_line:
            return 0
        else:
            intersect = (right_line - left_line) * (bottom_line - top_line)
            return (intersect / (sum_area - intersect))*1.0

    def recall_precision(self, gt, pre, iou_thr=0.5):
        # recall
        recall_sum = len(gt)
        recall_cnt = 0
        precision_sum = len(pre)
        precision_cnt = 0
        for g_bbox in gt:
            for p_bbox in pre:
                iou_ = self.compute_iou(g_bbox, p_bbox)
                # print('iou_of_recall: ', iou_)
                if iou_ >= iou_thr:
                    recall_cnt += 1
                    break
        for p_bbox in pre:
            for g_bbox in gt:
                iou_ = self.compute_iou(g_bbox, p_bbox)
                # print('iou_of_precision: ', iou_)
                if iou_ >= iou_thr:
                    precision_cnt += 1
                    break

        if recall_cnt > recall_sum:
            print(" ----------------------》》》 error recall")
        if precision_cnt > precision_sum:
            print(" ----------------------》》》 error precision")
        return (recall_sum, recall_cnt, precision_sum, precision_cnt)

    def eval_precisely(self):
        recall_sum, recall_cnt, precision_sum, precision_cnt = 0., 0., 0., 0.
        mAPs = []
        mrecall = []
        pic_cnt = 0

        tbar = tqdm(self.img_files)
        for i, identity in enumerate(tbar):
            # load ground truth
            im = Image.open(identity)
            im.verify()  # PIL verify
            shape = self.exif_size(im)
            img = cv2.imread(identity)
            w = shape[0]
            h = shape[1]

            annotation = self.label_files[i]
            bboxes, labels = self._load_yolo_boxes_and_labels(annotation, w, h)
            gt_list = bboxes

            # inference
            # inp = self.prep_func(identity)
            # out = self.model_(inp)
            # out = self.posp_func(out)
            out = self.infer_func(identity)

            for b in bboxes:
                x1, y1, x2, y2 = b
                cv2.rectangle(img, (x1, y1), (x2, y2), (0, 0, 255), 2)

            pre_list = []
            for m_ in out:
                x1, y1, x2, y2, cls_id, conf = m_
                x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)
                if conf >= self.conf_thr:
                    pre_list.append((x1, y1, x2, y2))
                    cv2.rectangle(img, (x1, y1), (x2, y2), (0, 255, 0), 2)

            pic_cnt += 1
            eval_list = self.recall_precision(
                gt_list, pre_list, iou_thr=self.iou_thr)
            # logger.info(eval_list)

            recall_sum += eval_list[0]
            recall_cnt += eval_list[1]
            precision_sum += eval_list[2]
            precision_cnt += eval_list[3]
            # print('{} {} {} {}'.format(recall_sum, recall_cnt, precision_sum, precision_cnt))
            if (recall_sum > 0) and (precision_sum > 0):
                if eval_list[2] > 0:
                    mAPs.append(eval_list[3]/eval_list[2])
                else:
                    if eval_list[0] > 0:
                        mAPs.append(0)
                if eval_list[0] > 0:
                    mrecall.append(eval_list[1]/eval_list[0])
                # print(recall_cnt,recall_sum,precision_cnt,precision_sum)

            tbar.set_postfix(recall='{:.3f}'.format(recall_cnt/recall_sum), precision='{:.3f}'.format(
                precision_cnt/precision_sum), mAP='{:.3f}'.format(np.mean(mAPs)), refresh=True)

            if self.hold_vis:
                cv2.imshow('aa', img)
                if cv2.waitKey(0) & 0xFF == ord('q'):
                    print('Pressed Q, continue eval without vis.')
                    self.hold_vis = False
                    cv2.destroyAllWindows()
        print("[\n{:6d}] conf_thr: {:.2f}, iou_thr: {:.2f}, recall: {:.5f}, precision: {:.5f}, mrecall: {:.3f}, map: {:.3f}".
              format(pic_cnt, self.conf_thr, self.iou_thr, recall_cnt/recall_sum, precision_cnt/precision_sum, np.mean(mrecall), np.mean(mAPs)))

    def eval(self):
        recall_sum, recall_cnt, precision_sum, precision_cnt = 0., 0., 0., 0.
        mAPs = []
        mrecall = []
        pic_cnt = 0

        for i, identity in enumerate(self.img_files):
            # load ground truth
            im = Image.open(identity)
            im.verify()  # PIL verify
            shape = self.exif_size(im)
            w = shape[0]
            h = shape[1]

            annotation = self.label_files[i]
            bboxes, labels = self._load_yolo_boxes_and_labels(annotation, w, h)
            gt_list = bboxes

            # inference
            inp = self.prep_func(identity)
            out = self.model_(inp)
            out = self.posp_func(out)

            pre_list = []
            for m_ in out:
                x1, y1, x2, y2, conf = m_
                x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)
                if conf >= self.conf_thr:
                    pre_list.append((y1, x1, y2, x2))

            pic_cnt += 1
            eval_list = self.recall_precision(
                gt_list, pre_list, iou_thr=self.iou_thr)

            recall_sum += eval_list[0]
            recall_cnt += eval_list[1]
            precision_sum += eval_list[2]
            precision_cnt += eval_list[3]
            if (recall_sum > 0) and (precision_sum > 0):
                if eval_list[2] > 0:
                    mAPs.append(eval_list[3]/eval_list[2])
                else:
                    if eval_list[0] > 0:
                        mAPs.append(0)
                if eval_list[0] > 0:
                    mrecall.append(eval_list[1]/eval_list[0])
                # print(recall_cnt,recall_sum,precision_cnt,precision_sum)
                print("  {:6d}>  -->>conf_thr :{:.2f} , iou_thr :{:.2f} , recall :{:.5f} , precision :{:.5f}, mrecall :{:.3f}, map :{:.3f}".
                      format(pic_cnt, self.conf_thr, self.iou_thr, recall_cnt/recall_sum, precision_cnt/precision_sum, np.mean(mrecall), np.mean(map)), end='\r')
            print('all recall: {}, all precision: {}'.format(
                recall_sum, precision_sum))



#
# Copyright (c) 2020 JinTian.
#
# This file is part of alfred
# (see http://jinfagang.github.io).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""

split txt file with ratios

alfred cab split -f all.txt -r 0.1,0.8,0.1 -n train,val,test

"""
import os
import glob
from alfred.utils.log import logger as logging
import numpy as np


def split_txt_file(f, ratios, names):
    assert os.path.exists(f), '{} not exist.'.format(f)
    if not ratios:
        ratios = [0.2, 0.8]
    else:
        ratios = [float(i) for i in ratios.split(',')]
    logging.info('split ratios: {}'.format(ratios))

    if not names:
        names = ['part_{}'.format(i) for i in range(len(ratios))]
    else:
        names = names.split(',')
    names = [i+'.txt' for i in names]
    logging.info('split save to names: {}'.format(names))

    a = sum(ratios)
    if a != 1.:
        logging.info(
            'ratios: {} does not sum to 1. you must change it first.'.format(ratios))
        exit(1)

    # read txt file
    with open(f, 'r') as f:
        lines = f.readlines()
        lines_no_empty = [i for i in lines if i != '' and i != '\n']
        logging.info('to split file have all {} lines. droped {} empty lines.'.format(len(lines),
                                                                                      len(lines) - len(lines_no_empty)))
        lines = lines_no_empty
        # split with ratios
        last_lines = 0
        for i, r in enumerate(ratios):
            one = lines[last_lines: last_lines+int(r * len(lines))]
            with open(names[i], 'w') as ff:
                ff.writelines(one)
                logging.info('Part {} saved into: {}. portion: {}/{}={}'.format(
                    i, names[i], len(one), len(lines), len(one)/(len(lines))))
            last_lines += len(one)
    logging.info('split done.')



import socket
import numpy as np


class LocalSocketExchanger:
    """
    send and receive data between 2 program
    """

    def __init__(self, ip="127.0.0.1", port=5005, is_server=True) -> None:
        self.MAX_BUFFER_SIZE = 20480

        self.ip = ip
        self.port = port
        self.did_received_data_func = None
        self.is_server = is_server
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._connect()

    def _connect(self):
        if self.is_server:
            self.socket.bind((self.ip, self.port))
            self.socket.listen(3)
            self.conn, addr = self.socket.accept()
            print('Server is ready. ', addr)
        else:
            self.socket.connect((self.ip, self.port))

    def send(self, data):
        """
        send any type of data, it might be List, NDArray, etc
        """
        if not self.is_server:
            self.socket.send(data)
        else:
            print('server can not send data for now.')

    def did_received_data(self, func):
        self.did_received_data_func = func

    def receive(self):
        data = self.conn.recv(self.MAX_BUFFER_SIZE)
        a = np.frombuffer(data, dtype=np.float32)
        if len(a) > 0:
            # print("receive data: ", a, a.shape)
            if self.did_received_data_func:
                self.did_received_data_func(a)

    def listen_and_loop(self):
        while True:
            self.receive()
